"""
File Purpose: Contour: 2D contours; movie of contours for 3D data.
"""
import matplotlib.pyplot as plt
#import matplotlib.ticker as mpl_ticker  # [TODO] use this for good default levels.
import matplotlib.lines as mpl_lines
import numpy as np
import xarray as xr

from .labels import xarray_title_plot_node
from .movies import MoviePlotElement, MoviePlotNode
from .plot_settings import PlotSettings
from .plot_tools import (
    update_vlims, set_margins_if_provided, set_lims_if_provided,
    PlotDimsMixin,
) 
from ..errors import PlottingAmbiguityError
from ..tools import (
    alias, alias_key_of, alias_child, simple_property,
    UNSET,
    pcAccessor,
    xarray_gaussian_filter,
    xarray_ensure_dims, xarray_as_array, xarray_fill_coords,
)


@PlotSettings.format_docstring()
class XarrayContourPlotElement(MoviePlotElement):
    '''plot contours of 2D data.
    
    array: xarray.DataArray, probably ndim=2.
        the data to be plotted.
    add_colorbar: {add_colorbar}
    colorbar_linewidth: {colorbar_linewidth}
    colorbar_linestyle: {colorbar_linestyle}
    colorbar_lines_set: {colorbar_lines_set}
    label: {label}
        UNSET --> no label.
    legend_handle_kw: {legend_handle_kw}

    contourf: {contourf}
    '''
    def __init__(self, array, *, ax=None,
                 contourf=PlotSettings.default('contourf'),
                 **kw_super):
        kw_super.update(contourf=contourf)
        super().__init__(**kw_super)
        self._ax_init = ax
        self.array = array
        self.init_contour()

    array = alias_key_of('data', 'array', doc='''array for plot. Internally, stored at self.data['array']''')
    contour = simple_property('_contour', doc='''contour plot (instance of matplotlib.contour.QuadContourSet)''')
    ax = alias_child('contour', 'axes', if_no_child=None, doc='''the axes containing this contour plot''')
    fig = alias_child('ax', 'figure', if_no_child=None, doc='''figure containing this contour plot''')

    # # # PLOTTING THIS ELEMENT # # #
    def _contour_plotter(self, array):
        '''return plotter to use for making contours.
        array.plot.contourf or array.plot.contour, depending on self.plot_settings['contourf'].
        '''
        return array.plot.contourf if self.plot_settings['contourf'] else array.plot.contour

    def _contour_plotter_kwargs(self):
        '''return kwargs to use for making contours.
        array.plot.contourf or array.plot.contour, depending on self.plot_settings['contourf'].
        '''
        fname = 'contourf' if self.plot_settings['contourf'] else 'contour'
        return self.plot_settings.get_mpl_kwargs(f'xarray.DataArray.plot.{fname}')

    def init_contour(self):
        '''initialize the contour plot; actually plot the data.
        stores plotted object in self.contour and returns self.contour.
        '''
        array = self.array
        # plot settings / kwargs / bookkeeping
        kw_plot = self._contour_plotter_kwargs()
        kw_plot = update_vlims(array, kw_plot)  # [TODO] and update levels too?
        kw_plot.update(add_colorbar=False)  # we handle colorbar via self.colorbar() instead.
        if 'label' in kw_plot:
            kw_plot['label'] = kw_plot['label'].format(**xarray_nondim_coords(array))
        # make plot
        plotter = self._contour_plotter(array)
        self.contour = plotter(ax=self._ax_init, **kw_plot)
        # formatting
        aspect = self.plot_settings['aspect']  # xarray plotter doesn't like 'aspect' kwarg; assign here instead.
        if aspect is not None:
            self.ax.set_aspect(aspect)
        set_margins_if_provided(self.plot_settings, ax=self.ax)
        set_lims_if_provided(self.plot_settings, ax=self.ax)
        # make colorbar if appropriate
        if self.plot_settings['add_colorbar']:
            self.colorbar()
        return self.contour

    # # # UPDATING (REQUIRED BY PARENT) # # #
    def update_data(self, data):
        '''update the plot using data['array'].
        return the list of all updated matplotlib Artist objects.
        '''
        # [TODO] is there a way to update existing contours instead of making new ones?
        array = data['array']
        # remove existing contours (if any). Remember some old values first.
        old = getattr(self, 'contour', None)
        assert old is not None, "ContourPlotElement.update_data called before init_contour?"
        ax_old = self.ax
        kw_old = dict(  # be sure to use the same limits as old.
                levels = old.levels,
                vmin = old.get_clim()[0],
                vmax = old.get_clim()[1],
                cmap = old.cmap,
                )
        old.remove()
        # plot new contours
        kw_plot = self._contour_plotter_kwargs()
        kw_plot.update(add_colorbar=False,   # updating data --> no new colorbar...
                        add_labels=False,    # --> no new labels (from here) either.
                        )
        kw_plot = update_vlims(array, kw_plot)  # [TODO] and update levels too?
        kw_plot.update(kw_old)
        plotter = self._contour_plotter(array)
        self.contour = plotter(ax=ax_old, **kw_plot)
        return [self.contour]

    # # # ADDING STUFF TO PLOT # # #
    def colorbar(self, *,
                 colorbar_linewidth=UNSET, colorbar_linestyle=UNSET, colorbar_lines_set=UNSET,
                 **kw_plt_colorbar):
        '''add a colorbar to the plot. sets self.cbar = Colorbar object, and returns it.'''
        contour = self.contour
        kw_plt_colorbar.update(self.plot_settings.get('colorbar_kw', default=dict()))
        self.cbar = self.fig.colorbar(contour, **kw_plt_colorbar)
        # formatting for colorbar lines
        clines = self.cbar.ax.get_children()[1]  # matplotlib.collections.LineCollection of lines on colorbar.
        kw_clines = self.plot_settings.get('colorbar_lines_set', colorbar_lines_set, default=dict())
        kw_clines = kw_clines.copy()  # copy <--> don't edit default.
        lw = self.plot_settings.get('colorbar_linewidth', colorbar_linewidth)
        ls = self.plot_settings.get('colorbar_linestyle', colorbar_linestyle)
        if ls is not None:
            kw_clines['linestyle'] = ls
        if lw is not None:
            if isinstance(lw, tuple) and len(lw)==2:
                lwmin, lwmax = lw
                lwnow = clines.get_linewidth()  # current linewidth (is a list or array, e.g. [2])
                lwuse = np.clip(lwnow, lwmin, lwmax)
                kw_clines['linewidth'] = lwuse
            else:
                kw_clines['linewidth'] = lw
        clines.set(**kw_clines)
        return self.cbar

    def legend_handle(self, label=UNSET, **kw_line_2D_set):
        '''return a handle to be used in a legend for this plot.
        Will look like the first contour line in self.contour,
            but then apply any settings from kw_line_2D_set.
        E.g. provide color='black' to make a line like the first contour, but black.

        Recommend:
            plt.legend(handles=[self.legend_handle(), ...])
        filling in the ... with any other artists who have legend elements to draw.
        To grab the default list of handles, you can use:
            default_handles = ax.get_legend_handles_labels()[0].
        where ax is the axis where objects are plotted (maybe self.ax, or plt.gca()).
        Putting these together would result in something like:
            plt.legend(handles=[self.legend_handle(), *default_handles], ...)
        If the handles aren't long enough to visually distinguish, try handlelength=5 or more.
        '''
        baseline = self.contour.legend_elements()[0][0]  # Line2D matching 0th contour style.
        result = mpl_lines.Line2D([0],[0])
        result.update_from(baseline)
        label = self.plot_settings.get('label', label)
        kw_line_2D_set.update(self.plot_settings.get('legend_handle_kw', default=dict()))
        result.set(**kw_line_2D_set, label=label)
        return result

    handle = alias('legend_handle')


@pcAccessor.register('contour')
@PlotSettings.format_docstring()
class XarrayContour(MoviePlotNode, PlotDimsMixin):
    '''MoviePlotNode of contours.
    stores an XarrayContourPlotElement & has methods for plotting & updating the plot.
    "contour" refers to a matplotlib.contour.QuadContourSet object, e.g. the result of plt.contour.

    Troubleshooting jagged contours? Try using blur=1 or larger!

    array: xarray.DataArray, probably ndim=3.
        the data to be plotted. if ndim=2, can still plot, but nothing to animate.
    t: None or str
        the array dimension which frames will index. E.g. 'time'.
        None -> infer from array & any other provided dimensions.
    x, y: None or str
        if provided, tells dimensions for the x, y plot axes.
        None -> infer from array & any other provided dimensions.
    ax: None or Axes
        the attached mpl.axes.Axes object.
        None --> will use self.ax=plt.gca() when getting self.ax for the first time.
    blur: None or number
        if non-None, apply gaussian filter to array before storing it;
        use sigma=blur, and only blur along x_plot_dim & y_plot_dim.
        e.g. equivalent to arr.pc.blur(['x', 'y'], sigma=blur).pc.contour(..., blur=None)
        This is useful if the contours are too jagged / noisy!

    init_plot: {init_plot}

    label: {label}
    legend_handle_kw: {legend_handle_kw}

    {kw_ax_margin}

    title: {title}
    title_font: {title_font}
    title_y: {title_y}
    title_kw: {title_kw}
    '''
    element_cls = XarrayContourPlotElement

    def __init__(self, array, t=None, *, x=None, y=None, ax=None, blur=None,
                 init_plot=PlotSettings.default('init_plot'),  # <- could go in kw_super, but explicit is nice.
                 label=PlotSettings.default('label'),
                 **kw_super):
        if isinstance(array, xr.Dataset):
            array = xarray_as_array(array)
        array = xarray_fill_coords(array, need=[x, y, t])
        array = xarray_ensure_dims(array, ({x, y, t} - {None}), promote_dims_if_needed=True)
        self.array = array
        self.t = t
        self.x = x
        self.y = y
        self._ax_init = ax
        kw_super.update(init_plot=init_plot, label=label)
        self.plot_dims_attrs = self.plot_dims_attrs.copy()  # copy cls attr to avoid overwriting by accident
        if blur is not None:
            self.array = xarray_gaussian_filter(self.array, self.xy_plot_dims, sigma=blur)
        super().__init__(**kw_super)

    # # # PROPERTIES # # #
    ax = alias_child('obj', 'ax', doc='''mpl.axes.Axes where this Contour is plotted.''')
    @ax.getter
    def ax(self):  # ax=None if not self.plotted
        return self.obj.ax if self.plotted else None

    fig = alias_child('obj', 'fig', doc='''figure where this Contour is plotted.''')
    @fig.getter
    def fig(self):  # fig=None if not self.plotted
        return self.obj.fig if self.plotted else None

    handle = alias_child('obj', 'handle', doc='''creates legend handle to use for this Contour.''')

    # # # PLOTTING METHODS (REQUIRED BY PARENT CLASS) # # #
    def init_plot(self):
        '''plot for the first time. Save the XarrayContourPlotElement at self.obj'''
        self._init_plot_checks()
        frame = self.plot_settings['init_plot_frame']
        data = self.get_data_at_frame(frame)
        # get settings for plot
        kw_plot = self.plot_settings.get_mpl_kwargs('pc.XarrayContourPlotElement')
        # -- determine vlims (based on the entire array, not just this frame)
        kw_plot = update_vlims(self.array, kw_plot)
        # >> actually plot the contour <<
        self.obj = self.element_cls(data['array'], ax=self._ax_init, **kw_plot)
        # -- if add_labels, add title:
        if self.plot_settings['add_labels']:
            self.plot_title()
        # bookkeeping
        self.frame = frame

    # [TODO] encapsulate similarities between this & XarrayImage???
    def get_data_at_frame(self, frame):
        '''returns {'array': array at this frame}, properly transposed & ready for plotting.'''
        t = self.t_plot_dim
        result = self.array
        if t is not None:  # there is a dimension for the time axis.
            result = result.isel({t: frame})
        x, y = self.xy_plot_dims   # (x and y are strings, the dimension names)
        result = result.transpose(y, x)   # transposed to get correct order for plotting
        return {'array': result}

    def get_nframes_here(self):
        '''return the number of frames that could be in the movie, based on this node.'''
        t = self.t_plot_dim
        if t is None:
            return 1
        return len(self.array.coords[t])

    # # # INFERRING DIMS (REQUIRED BY PARENT CLASS) # # #
    plot_dims_attrs = dict(x='x', y='y', t='t', dims='_array_dims', t_necessary_if='_array_is_not_2D')
    _array_dims = property(lambda self: self.array.dims, doc='''the dimensions of the array''')
    # x_plot_dim, y_plot_dim, t_plot_dim are already defined by PlotDimsMixin.

    def _array_is_not_2D(self):
        '''tells whether self.array is not 2D'''
        return self.array.ndim != 2

    def _has_valid_plot_dims(self):
        '''return whether this array is plottable, i.e. all plot_dims can be inferred,
        and array doesn't have more than 3 dims.'''
        if self.array.ndim > 3:
            return False
        try:
            _plot_dims = self.xyt_plot_dims
        except PlottingAmbiguityError:
            return False
        return True

    # # # TITLE # # #
    def plot_title(self):
        '''adds title (as a MovieTextNode) on self.ax.
        raise PlottingAmbiguityError if title already plotted
            (this prevents having multiple title nodes).
        '''
        if hasattr(self, 'title_node'):
            raise PlottingAmbiguityError(f'{type(self).__name__} title was already plotted!')
        title = self.plot_settings['title']
        if title is None:
            plt.title('')  # remove xarray's default title.
        else:
            kw_title = {**self.plot_settings.kw, 'init_plot': True}
            title_node = xarray_title_plot_node(self.array, title, t=self.t_plot_dim,
                                                ax=self.ax, parent=self, **kw_title)
            self.title_node = title_node  # <-- for bookkeeping/debugging



"""
File Purpose: Fluid, FluidList, FluidDimension, FluidHaver, jFluidDimension, jFluidHaver, FluidsHaver
"""
import xarray as xr

from .dimension_tools import (
    DimensionValue, DimensionValueList, UniqueDimensionValue,
    DimensionSpecialValueSpecifier, DimensionSingleValueSpecifier,
    Dimension, DimensionHaver,
)
from ..errors import FluidKeyError, FluidValueError, InputError
from ..tools import (
    alias, elementwise_property,
    xarray_rename,
    is_iterable,
    UNSET,
    Sentinel,
)


''' --------------------- Fluid & FluidList --------------------- '''

class Fluid(DimensionValue):
    '''fluid... unchanging properties of the fluid.

    name: the name (str) of the fluid. if None, cannot convert to str.
    i: the index (int) of the fluid (within a FluidList). if None, cannot convert to int.

    other inputs should be in "elementary" units, i.e.:
        m: mass, in atomic mass units  (1 for H+)
        q: charge, in elementary charge units  (1 for H+; -1 for e-)
    '''
    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'name', 'i', 'm', 'q'}

    def __init__(self, name=None, i=None, *, m=None, q=None):
        super().__init__(name, i)
        self.m = m
        self.q = q

    name = alias('s')

    def _repr_contents(self):
        '''contents used by self.__repr__'''
        contents = [repr(val) for val in [self.name, self.i] if val is not None]
        if self.m is not None:
            contents.append(f'm={self.m:.3f}')
        if self.q is not None:
            contents.append(f'q={self.q}')
        return contents

    def __repr__(self):
        contents = self._repr_contents()
        return f'{type(self).__name__}({", ".join(contents)})'

    def is_neutral(self):
        '''tells whether self is a neutral fluid. i.e. self.q == 0.
        if self.q is None, returns None instead of bool.
        '''
        if getattr(self, 'q', None) is None: return None
        return self.q == 0

    def is_ion(self):
        '''tells whether self is an ion. i.e. self.q > 0.
        if self.q is None, returns None instead of bool.
        '''
        if getattr(self, 'q', None) is None: return None
        return self.q > 0

    def is_electron(self):
        '''tells whether self is an electron. i.e. self.q < 0.
        if self.q is None, returns None instead of bool.
        '''
        if getattr(self, 'q', None) is None: return None
        return self.q < 0

    def is_charged(self):
        '''tells whether self is a charged fluid. i.e. self.q != 0.
        if self.q is None, returns None instead of bool.
        '''
        if getattr(self, 'q', None) is None: return None
        return self.q != 0


class FluidList(DimensionValueList):
    '''list of fluids'''
    _dimension_key_error = FluidKeyError
    _dimension_value_error = FluidValueError
    value_type = Fluid

    name = elementwise_property('name')
    m = elementwise_property('m')
    q = elementwise_property('q')

    def is_neutral(self):
        '''return xarray.DataArray telling whether each fluid in self is neutral.'''
        return xr.DataArray([fluid.is_neutral() for fluid in self], coords=self)
    def is_ion(self):
        '''return xarray.DataArray telling whether each fluid in self is an ion.'''
        return xr.DataArray([fluid.is_ion() for fluid in self], coords=self)
    def is_electron(self):
        '''return xarray.DataArray telling whether each fluid in self is an electron.'''
        return xr.DataArray([fluid.is_electron() for fluid in self], coords=self)
    def is_charged(self):
        '''return xarray.DataArray telling whether each fluid in self is charged.'''
        return xr.DataArray([fluid.is_charged() for fluid in self], coords=self)

    def i_neutrals(self):
        '''return indices of neutrals in self'''
        return [i for i, fluid in enumerate(self) if fluid.q == 0]
    def i_ions(self):
        '''return indices of ions in self'''
        return [i for i, fluid in enumerate(self) if fluid.q > 0]
    def i_electrons(self):
        '''return indices of electrons in self'''
        return [i for i, fluid in enumerate(self) if fluid.q < 0]
    def i_charged(self):
        '''return indices of charged fluids in self'''
        return [i for i, fluid in enumerate(self) if fluid.q != 0]

    def i_neutral(self):
        '''return index of the single neutral fluid from self. Crash if there isn't exactly 1.
        if 0, raise FluidKeyError. if 2+, raise FluidValueError.
        '''
        neutrals = self.i_neutrals()
        if len(neutrals) == 0:
            raise self._dimension_key_error('expected exactly 1 neutral fluid, but got 0')
        elif len(neutrals) >= 2:
            raise self._dimension_value_error(f'expected exactly 1 neutral fluid, but got {len(neutrals)}')
        return neutrals[0]

    def i_ion(self):
        '''return index of the single ion fluid from self. Crash if there isn't exactly 1.
        if 0, raise FluidKeyError. if 2+, raise FluidValueError.
        '''
        ions = self.i_ions()
        if len(ions) == 0:
            raise self._dimension_key_error('expected exactly 1 ion fluid, but got 0')
        elif len(ions) >= 2:
            raise self._dimension_value_error(f'expected exactly 1 ion fluid, but got {len(ions)}')
        return ions[0]

    def i_electron(self):
        '''return index of the single electron fluid from self. Crash if there isn't exactly 1.
        if 0, raise FluidKeyError. if 2+, raise FluidValueError.
        '''
        electrons = self.i_electrons()
        if len(electrons) == 0:
            raise self._dimension_key_error('expected exactly 1 electron fluid, but got 0')
        elif len(electrons) >= 2:
            raise self._dimension_value_error(f'expected exactly 1 electron fluid, but got {len(electrons)}')
        return electrons[0]

    def neutrals(self):
        '''return FluidList of neutrals from self.'''
        return self[self.i_neutrals()]
    def ions(self):
        '''return FluidList of ions from self.'''
        return self[self.i_ions()]
    def electrons(self):
        '''return FluidList of electrons from self.'''
        return self[self.i_electrons()]
    def charged(self):
        '''return FluidList of charged fluids from self.'''
        return self[self.i_charged()]

    def get_neutral(self):
        '''return the single neutral fluid from self. Crash if there isn't exactly 1.
        if 0, raise FluidKeyError. if 2+, raise FluidValueError.
        '''
        return self[self.i_neutral()]
    def get_ion(self):
        '''return the single ion fluid from self. Crash if there isn't exactly 1.
        if 0, raise FluidKeyError. if 2+, raise FluidValueError.
        '''
        return self[self.i_ion()]
    def get_electron(self):
        '''return the single electron fluid from self. Crash if there isn't exactly 1.
        if 0, raise FluidKeyError. if 2+, raise FluidValueError.
        '''
        return self[self.i_electron()]


class UniqueFluid(UniqueDimensionValue, Fluid):
    '''unique fluid; Sentinel value with Fluid type.'''
    pass


SINGLE_FLUID = UniqueFluid('SINGLE_FLUID')  # to indicate "get single-fluid value"


''' --------------------- Convenient DimensionSpecialValueSpecifier objects --------------------- '''

class FluidSpecialValueSpecifier(Sentinel, DimensionSpecialValueSpecifier):
    '''class to specify special values for FluidDimension.
    E.g. fluids.get(ELECTRON) is equivalent to fluids.get_electron().
    E.g. (fluid_dim.v = IONS) is equivalent to (fluid_dim.v = fluid_dim.values.get_electron())

    getter: str
        self specifies to use the value(s): fluids.{getter}().
    name: str
        name of this object; will only be used in repr.
    iseler: None or str
        self specifies to use (during xarray_isel or xarray_sel) indexes: fluids.{iseler}()

    additional args & kw go to Sentinel.__new__
    '''
    isel_cls = FluidList  # if used in xarray_isel or xarray_sel, convert coord vals to this type first.

    def __new__(cls, getter, name=None, *args_super, iseler=None, **kw_super):
        return super().__new__(cls, name, *args_super, **kw_super)  # Sentinel.__new__
        # python auto-calls __init__ with these args, afterwards.
        # specifier-related logic is all inherited from DimensionSpecialValueSpecifier

class FluidSingleValueSpecifier(Sentinel, DimensionSingleValueSpecifier):
    '''class to specify single values for FluidDimension.
    E.g. fluids.get(ION) is equivalent to fluids.get_ion().
    E.g. (fluid_dim.v = ION) is equivalent to (fluid_dim.v = fluid_dim.values.get_ion())

    getter: str
        self specifies to use the value(s): fluids.get_{getter}().
    name: str
        name of this object; will only be used in repr.
    iseler: None or str
        self specifies to use (during xarray_isel or xarray_sel) indexes: fluids.{iseler}()

    additional args & kw go to Sentinel.__new__
    '''
    isel_cls = FluidList  # if used in xarray_isel or xarray_sel, convert coord vals to this type first.

    def __new__(cls, getter, checker, name=None, *args_super, iseler=None, **kw_super):
        return super().__new__(cls, name, *args_super, **kw_super)  # Sentinel.__new__
        # python auto-calls __init__ with these args, afterwards.
        # specifier-related logic is all inherited from DimensionSingleValueSpecifier

CHARGED = FluidSpecialValueSpecifier('charged', iseler='i_charged', name='CHARGED')
ELECTRON = FluidSingleValueSpecifier('get_electron', 'is_electron', iseler='i_electron', name='ELECTRON')
ELECTRONS = FluidSpecialValueSpecifier('electrons', iseler='i_electrons', name='ELECTRONS')
ION = FluidSingleValueSpecifier('get_ion', 'is_ion', iseler='i_ion', name='ION')
IONS = FluidSpecialValueSpecifier('ions', iseler='i_ions', name='IONS')
NEUTRAL = FluidSingleValueSpecifier('get_neutral', 'is_neutral', iseler='i_neutral', name='NEUTRAL')
NEUTRALS = FluidSpecialValueSpecifier('neutrals', iseler='i_neutrals', name='NEUTRALS')


''' --------------------- FluidDimension, FluidHaver --------------------- '''

class FluidDimension(Dimension, name='fluid', plural='fluids',
                     value_error_type=FluidValueError, key_error_type=FluidKeyError):
    '''fluid dimension, representing current value AND list of all possible values.
    Also has various helpful methods for working with this Dimension.
    '''
    pass  # behavior inherited from Dimension.


@FluidDimension.setup_haver
class FluidHaver(DimensionHaver, dimension='fluid', dim_plural='fluids'):
    '''class which "has" a FluidDimension. (FluidDimension instance will be at self.fluid_dim)
    self.fluid stores the current fluid (possibly multiple). If None, use self.fluids instead.
    self.fluids stores "all possible fluids" for the FluidHaver.
    Additionally, has various helpful methods for working with the FluidDimension,
        e.g. current_n_fluid, iter_fluids, take_fluid.
        See FluidDimension.setup_haver for details.
    '''
    def __init__(self, *, fluid=None, fluids=None, **kw):
        super().__init__(**kw)
        if fluids is not None: self.fluids = fluids
        self.fluid = fluid


''' --------------------- jFluidDimension, jFluidHaver --------------------- '''        

class jFluidDimension(Dimension, name='jfluid', plural='jfluids',
                      value_error_type=FluidValueError, key_error_type=FluidKeyError):
    '''jfluid dimension, representing current value AND list of all possible values.
    Also has various helpful methods for working with this Dimension.
    '''
    pass  # behavior inherited from Dimension.


@jFluidDimension.setup_haver
class jFluidHaver(DimensionHaver, dimension='jfluid', dim_plural='jfluids'):
    '''class which "has" a jFluidDimension. (jFluidDimension instance will be at self.jfluid_dim)
    self.jfluid stores the current jfluid (possibly multiple). If None, use self.fluids instead.
    self.jfluids stores "all possible jfluids" for the jFluidHaver.
    Additionally, has various helpful methods for working with the FluidDimension,
        e.g. current_n_fluid, iter_fluids, take_fluid.
        See jFluidDimension.setup_haver for details.

    (Some variables, like 'nusj' depend on multiple fluids; for those variables use fluid and jfluid.)
    '''
    def __init__(self, *, jfluid=None, jfluids=None, **kw):
        super().__init__(**kw)
        if jfluids is not None: self.jfluids = jfluids
        self.jfluid = jfluid

    def getj(self, var, *args__get, jfluid=UNSET, **kw__get):
        '''returns self(var), but for jfluid instead of fluid.
        jfluid: UNSET, None, or any jfluid specifier.
            if provided, use this instead of self.jfluid.

        Example:
                m_s = self('m')
                with self.using(fluids=self.jfluids, fluid=self.jfluid):
                    m_j_0 = self('m')
                m_j_1 = self.getj('m')
            here, the values stored in the variables will be:
                m_s = mass of self.fluid
                m_j_0 = mass of self.jfluid. But, fluid dimension will be labeled 'fluid'
                m_j_1 = mass of self.jfluid. fluid dimension will be labeled 'jfluid'.

        additional args and kwargs are passed to self(var)
        '''
        with self.using(fluids=self.jfluids, fluid=self.jfluid if jfluid is UNSET else jfluid):
            result = self(var, *args__get, **kw__get)
        if not kw__get.get('item', False):  # didn't do result.item(); i.e. result is still an xarray
            result = xarray_rename(result, {'fluid': 'jfluid'})
        return result


''' --------------------- FluidsHaver --------------------- '''

class FluidsHaver(FluidHaver, jFluidHaver):
    '''class which "has" a FluidDimension and a jFluidDimension.
    Most FluidHavers will probably be better off as FluidsHavers,
        since having a jFluid enables to calculate stuff like collision frequencies. 
    '''
    def get_neutral(self, var=None, *args__get, **kw__get):
        '''returns self(var), but for the neutral fluid.
        if var is None, instead returns the neutral fluid itself.

        if there is exactly 1 neutral fluid in self.fluids, returns self(var, fluid=neutral_fluid).
        otherwise, if there is exactly 1 in self.jfluids, returns self.getj(var, jfluid=neutral_fluid).
        otherwise (0 or 2+ neutral fluids), crash:
            if 0 neutral fluids anywhere, raise FluidKeyError.
            if 2+ neutral fluids, raise FluidValueError.
        '''
        # get neutral fluid
        neutrals = self.fluids.neutrals()
        if len(neutrals) == 1:
            neutral_fluid = neutrals[0]
            from_j = False
        else:  # need to check jfluids.
            neutralj = self.jfluids.neutrals()
            if len(neutralj) == 1:
                neutral_fluid = neutralj[0]
                from_j = True
            else:  # didn't find exactly 1 fluid in self.fluid nor self.jfluids.
                nfound = len(neutrals) + len(neutralj)
                if nfound > 0:
                    errmsg = ('expected exactly 1 neutral fluid in self.fluids or self.jfluids, '
                              f'but found {len(neutrals)} in fluids and {len(neutralj)} in jfluids.')
                    raise FluidValueError(errmsg)
                else:  # nfound == 0
                    errmsg = 'expected 1 neutral fluid, but none found in self.fluids or self.jfluids.'
                    raise FluidKeyError(errmsg)
        # get result
        if var is None:
            return neutral_fluid
        if from_j:
            return self.getj(var, jfluid=neutral_fluid, *args__get, **kw__get)
        else:
            return self(var, fluid=neutral_fluid, *args__get, **kw__get)

    def _as_single_fluid_or_jfluid(self, fluid_or_jfluid):
        '''return the single fluid or jfluid corresponding to this input.
        fluid_or_jfluid: Fluid, str, int, or other fluid specifier
            attempts self._as_single_fluid(fluid_or_jfluid).
            if FluidKeyError, do self._as_single_jfluid(fluid_or_jfluid) instead.
            (if both fail, raise FluidKeyError.)
        (similar to _as_single_jfluid_or_fluid, but here looks in self.fluids first.)
        '''
        try:
            return self._as_single_fluid(fluid_or_jfluid)
        except FluidKeyError:
            return self._as_single_jfluid(fluid_or_jfluid)

    def _as_single_jfluid_or_fluid(self, jfluid_or_fluid):
        '''return the single jfluid or fluid corresponding to this input.
        jfluid_or_fluid: Fluid, str, int, or other fluid specifier
            attempts self._as_single_jfluid(jfluid_or_fluid).
            if FluidKeyError, do self._as_single_fluid(jfluid_or_fluid) instead.
            (if both fail, raise FluidKeyError.)
        (similar to _as_single_fluid_or_jfluid, but here looks in self.jfluids first.)
        '''
        try:
            return self._as_single_jfluid(jfluid_or_fluid)
        except FluidKeyError:
            return self._as_single_fluid(jfluid_or_fluid)

    def _get_fluid_or_jfluid_like(self, aliases, *, as_list=False, check_first='fluids'):
        '''return the fluid or jfluid corresponding to any of the aliases here.
        alisaes: list of str, int, Fluid, or other fluid specifier.
            each alias will be checked in self.fluids and self.jfluid.
        as_list: bool
            whether to return list in case of 0 or 2+ matches.
            False --> crash. FluidKeyError if 0 matches; FluidValueError if 2+ matches.
        check_first: 'fluids' or 'jfluids'
            which to check first: 'fluids' or 'jfluids'.
        '''
        if isinstance(aliases, str) or (not is_iterable(aliases)):
            raise InputError(f'aliases should be an iterable of fluid specifiers, but got aliases={aliases!r}.')
        if check_first == 'fluids':
            check0 = self._as_single_fluid
            check1 = self._as_single_jfluid
        elif check_first == 'jfluids':
            check0 = self._as_single_jfluid
            check1 = self._as_single_fluid
        else:
            raise InputError(f'check_first={check_first!r}. Expected "fluids" or "jfluids".')
        matches = set()  # fluids are hashable so this works!
        for a in aliases:
            # check 0
            try:
                match = check0(a)
            except FluidKeyError:
                pass  # this alias has no matches here; ignore it.
            else:
                matches.add(match)
            # check 1
            try:
                match = check1(a)
            except FluidKeyError:
                pass
            else:
                matches.add(match)
        if as_list:
            return list(matches)
        elif len(matches) == 1:
            return matches.pop()
        else:
            if len(matches) == 0:
                raise FluidKeyError(f'no matches found for aliases={aliases}.')
            else:
                raise FluidValueError(f'2+ matches ({matches}) found for aliases={aliases}')



"""
File Purpose: tools for simple properties
"""
import contextlib
import weakref

import numpy as np
import xarray as xr

from .docs_tools import format_docstring
from .sentinels import NO_VALUE, UNSET
from ..errors import InputError, InputConflictError

''' --------------------- Alias an attribute --------------------- '''

def alias(attribute_name, doc=None):
    '''returns a property which is an alias to attribute_name.
    if doc is None, use doc=f'alias to {attribute_name}'.
    '''
    return property(lambda self: getattr(self, attribute_name),
                    lambda self, value: setattr(self, attribute_name, value),
                    doc=f'''alias to {attribute_name}''' if doc is None else doc)

def alias_to_result_of(attribute_name, doc=None):
    '''returns a property which is an alias to the result of calling attribute_name.
    if doc is None, use doc=f'alias to {attribute_name}()'.
    '''
    return property(lambda self: getattr(self, attribute_name)(),
                    doc=f'''alias to {attribute_name}()''' if doc is None else doc)

def alias_child(child_name, attribute_name, doc=None, *, if_no_child=NO_VALUE, if_special_child=UNSET):
    '''returns a property which is an alias to obj.child_name.attribute_name.
    if doc is None, use doc=f'alias to self.{child_name}.{attribute_name}'.
    includes getter AND setter methods.

    if_no_child: NO_VALUE or any object
        if provided (not NO_VALUE), return this instead if child is None or doesn't exist.
    if_special_child: UNSET or dict
        if provided, dict of {special key: value to return if child == special key}
    '''
    if (if_no_child is NO_VALUE) and (if_special_child is UNSET):
        def getter(self):
            return getattr(getattr(self, child_name), attribute_name)
    elif (if_no_child is not NO_VALUE) and (if_special_child is UNSET):
        def getter(self):
            child = getattr(self, child_name, None)
            return if_no_child if child is None else getattr(child, attribute_name)
    elif (if_no_child is NO_VALUE) and (if_special_child is not UNSET):
        def getter(self):
            child = getattr(self, child_name, None)
            if child in if_special_child:
                return if_special_child[child]
            else:
                return getattr(child, attribute_name)
    elif (if_no_child is not NO_VALUE) and (if_special_child is not UNSET):
        def getter(self):
            child = getattr(self, child_name, None)
            if child is None:
                return if_no_child
            elif child in if_special_child:
                return if_special_child[child]
            else:
                return getattr(child, attribute_name)
    def setter(self, value):
        setattr(getattr(self, child_name), attribute_name, value)
    if doc is None:
        doc = f'''alias to self.{child_name}.{attribute_name}'''
    return property(getter, setter, doc=doc)

def alias_key_of(dict_attribute_name, key, *, default=NO_VALUE, setdefault_value=NO_VALUE, doc=None):
    '''returns a property which is an alias to obj.dict_attribute_name[key].
    if doc is None, use doc=f'alias to self.{dict_attribute_name}[{key!r}]'.
    includes getter, setter, and deleter methods.

    if setting value to UNSET, delete the key from the dict instead.

    default: any object
        if provided (not NO_VALUE), then getter returns self.dict_attribute_name.get(key, default).
    setdefault_value: any object
        if provided (not NO_VALUE), then getter returns self.dict_attribute_name.setdefault(key, setdefault_value).
    '''
    if default is not NO_VALUE and setdefault_value is not NO_VALUE:
        raise InputConflictError('cannot provide both default and setdefault_value.')
    if default is not NO_VALUE:
        def getter(self):
            return getattr(self, dict_attribute_name).get(key, default)
    elif setdefault_value is not NO_VALUE:
        def getter(self):
            return getattr(self, dict_attribute_name).setdefault(key, setdefault_value)
    else:
        def getter(self):
            return getattr(self, dict_attribute_name)[key]
    def deleter(self):
        del getattr(self, dict_attribute_name)[key]
    def setter(self, value):
        if value is UNSET:
            deleter(self)
        else:
            getattr(self, dict_attribute_name)[key] = value
    doc = f'''alias to self.{dict_attribute_name}[{key!r}]''' if doc is None else doc
    return property(getter, setter, deleter, doc=doc)

def alias_in(cls, attr, new_attr):
    '''sets cls.new_attr to be an alias for cls.attr.'''
    setattr(cls, new_attr, alias(attr))


''' --------------------- Other properties --------------------- '''

def weakref_property_simple(internal_attr, doc=None):
    '''defines a property which behaves like the value it contains, but is actually a weakref.
    stores internally at internal_attr. Also set self.{internal_attr}_is_weakref = True.
    Setting the value actually creates a weakref. Getting the value calls the weakref.
    Note: if failing to create weakref due to TypeError,
        (e.g. "TypeError: cannot create weak reference to 'int' object")
        then just store the value itself. Also set self.{internal_attr}_is_weakref = False.
    '''
    internal_attr_is_weakref = f'{internal_attr}_is_weakref'
    @format_docstring(attr=internal_attr)
    def get_attr(self):
        '''gets self.{attr}(), or just self.{attr} if not self.{attr}_is_weakref.'''
        self_attr = getattr(self, internal_attr)
        if getattr(self, internal_attr_is_weakref):
            return self_attr()
        else:
            return self_attr

    @format_docstring(attr=internal_attr)
    def set_attr(self, val):
        '''sets self.{attr}'''
        try:
            result = weakref.ref(val)
        except TypeError:
            setattr(self, internal_attr_is_weakref, False)
            result = val
        else:
            setattr(self, internal_attr_is_weakref, True)
        setattr(self, internal_attr, result)

    @format_docstring(attr=internal_attr)
    def del_attr(self):
        '''deletes self.{attr}'''
        delattr(self, internal_attr)
    
    return property(get_attr, set_attr, del_attr, doc)

def simple_property(internal_name, *, doc=None,
                    default=NO_VALUE, setdefault=NO_VALUE, setdefaultvia=NO_VALUE,
                    setable=True, delable=True,
                    valid=UNSET, validate_from=UNSET, validate_default=False):
    '''return a property with a setter and getter method for internal_name.

    default: NO_VALUE or any object
        if provided, getter uses this default if this attr has not been set.
    setdefault: NO_VALUE or callable of 0 arguments
        if provided, getter uses setdefault() if this attr has not been set.
        (if internal_name is unset, getter sets self.{internal_name} = setdefault(),
            then returns value at self.{internal_name}. Does not recalculate each time.)
        E.g. setdefault = dict, creates a new dict.
            (using default=dict() would share the dict amongst all instances;
            using setdefault=dict creates a new dict for each instance when needed.)
    setdefaultvia: NO_VALUE or str
        if provided, getter uses self.{setdefaultvia}() if attr has not been set.
        (if internal_name is unset, getter sets self.{internal_name} = self.{setdefaultvia}(),
            then returns value at self.{internal_name}. Does not recalculate each time.)
        E.g. setdefault = '_default_val' --> getter calls & returns self._default_val(),
            and remembers result, saving it to self.{internal_name}.

    setable: bool
        whether to allow this attribute to be set (i.e., define fset.)
    delable: bool
        whether to allow this attribute to be set (i.e., define fdel.)

    valid: UNSET or iterable
        if provided, only allow setting to values in valid.
    validate_from: UNSET or str
        if provided, get list of valid values from getattr(self, validate_from).
    validate_default: bool
        whether to apply validation to the results of setdefault() or setdefaultvia().
        disabled by default, for efficiency.
    '''
    # bookkeeping on inputs
    defaults = dict(default=default, setdefault=setdefault, setdefaultvia=setdefaultvia)
    provided_defaults = {key: val for key, val in defaults.items() if val is not NO_VALUE}
    if len(provided_defaults) > 1:
        raise InputConflictError(f'cannot provide more than 1 default, but got: {provided_defaults}')
    validation = dict(valid=valid, validate_from=validate_from)
    provided_validation = {key: val for key, val in validation.items() if val is not UNSET}
    if len(provided_validation) > 1:
        raise InputConflictError(f'cannot provide more than 1 validation kwarg, but got: {provided_validation}')
    if (not setable) and (provided_validation):
        raise InputConflictError(f'cannot provide validation kwarg if setable=False, but got: {provided_validation}')
    # getter
    if default is NO_VALUE and setdefault is NO_VALUE and setdefaultvia is NO_VALUE:
        def getter(self):
            return getattr(self, internal_name)
    elif default is not NO_VALUE:
        def getter(self):
            return getattr(self, internal_name, default)
    else:  # provided setdefault or setdefaultvia
        def getter(self):
            try:
                return getattr(self, internal_name)
            except AttributeError:
                pass  # handled below, to avoid stacked error message in case setdefault/via fails.
            # get default value:
            if setdefault is not NO_VALUE:
                result = setdefault()
            else:  # provided setdefaultvia
                defaultvia = getattr(self, setdefaultvia)
                result = defaultvia()
            # validate if relevant
            if validate_default and (valid is not UNSET or validate_from is not UNSET):
                allowed = getattr(self, validate_from) if valid is UNSET else valid
                if result not in allowed:
                    raise ValueError(f'setdefault gave {internal_name}={result!r}. Expected one of: {list(allowed)!r}')
            # set & return default value
            setattr(self, internal_name, result)
            return result
    # setter
    if setable:
        if (valid is UNSET and validate_from is UNSET):
            def setter(self, value):
                setattr(self, internal_name, value)
        else:
            def setter(self, value):
                allowed = getattr(self, validate_from) if valid is UNSET else valid
                if value not in allowed:
                    raise InputError(f'{internal_name}={value!r}. Expected one of: {list(allowed)!r}')
                setattr(self, internal_name, value)
    else:
        setter = None
    # deleter
    if delable:
        def deleter(self):
            delattr(self, internal_name)
    else:
        deleter = None
    return property(getter, setter, deleter, doc=doc)

def simple_tuple_property(*internal_names, doc=None, default=NO_VALUE):
    '''return a property which refers to a tuple of internal names.
    if 'default' provided (i.e., not NO_VALUE):
        - getter will have this default, if attr has not been set.
        - setter will do nothing if value is default.
        This applies to each name in internal_names, individually.
    '''
    if default is NO_VALUE:
        def getter(self):
            return tuple(getattr(self, name) for name in internal_names)
        def setter(self, value):
            for name, val in zip(internal_names, value):
                setattr(self, name, val)
    else:
        def getter(self):
            return tuple(getattr(self, name, default) for name in internal_names)
        def setter(self, value):
            for name, val in zip(internal_names, value):
                if val is not default:
                    setattr(self, name, val)
    def deleter(self):
        for name in internal_names:
            delattr(self, name)
    return property(getter, setter, deleter, doc=doc)

def elementwise_property(attr, *, as_array=True, array_dim=None, doc=None, default=NO_VALUE):
    '''return property which returns tuple(element.attr for element in self).

    as_array: bool or 'xarray', default True
        return np.array of result instead of list.
    array_dim: None or str
        if provided and as_array=True, return xarray.DataArray instead,
            with this dimension name, and coord = elements from self.
    doc: None or str
        the docstring for this property
    default: any value, default NO_VALUE
        if provided, use this for any element missing the 'attr' attribute, instead of crashing.

    The property also supports setting values, e.g.:
        self.attr = ['a', 'b', 'c', 'd'] sets self[0].attr='a', self[1].attr='b', etc.
        self.attr = 'common_val' sets self[0].attr='common_val', self[1].attr='common_val', etc.
    NOTE: to avoid ambiguity, the only non-"common" values are lists and tuples;
        all other types will be treated as "common" (hence, set the same value for each element).

    This property does not support deleting values.
    '''
    def getter(self):
        if default is NO_VALUE:
            result = tuple(getattr(el, attr) for el in self)
        else:
            result = tuple(getattr(el, attr, default) for el in self)
        if as_array:
            if array_dim is None:
                result = np.array(result)
            else:
                result = xr.DataArray(list(result), dims=[array_dim], coords={array_dim: self})
        return result
    def setter(self, value):
        if isinstance(value, (list, tuple)):
            for el, val in zip(self, value):
                setattr(el, attr, val)
        else:
            for el in self:
                setattr(el, attr, value)
    return property(getter, setter, doc=doc)


''' --------------------- Helpful to use with properties --------------------- '''

def dict_with_defaults_property(internal_name, *, key_aliases, doc=None):
    '''return a property which gives a dictionary, and gives key_aliases a chance to edit the dict.
    internal_name: str
        the name where this property is stored internally for obj.
    key_aliases: str or iterable of strs.
        the attributes of obj which are alias_key_of this property.
        str --> use obj.key_aliases.
    doc: None or str
        the docstring for this property
        if None, use f'dict with defaults determined by self.{key_aliases}'.

    Example:
        class Foo():
            special = dict_with_defaults_property('_special', key_aliases='_special_key_aliases')
            _special_key_aliases = ['special_key1', 'special_key2']
            special_key1 = alias_key_of('_special', 'key1', setdefault_value=5)
            special_key2 = alias_key_of('_special', 'key2')
        foo = Foo()
        # every time getting the dict from foo, use all the setdefault values as appropriate:
        foo.special  # --> dict(key1=5)
        foo.special = dict(key1=100)
        foo.special  # --> dict(key1=100)
        foo.special = dict(key3=30)
        foo.special  # --> dict(key3=30, key1=5)
        foo.special_key1 = 70
        foo.special  # --> dict(key3=30, key1=70)
        del foo.special['key1']
        foo.special  # --> dict(key3=30, key1=5)  # key1 was missing, so its setdefault value is used.

        # note that the dict itself is a normal dict; the magic happens when doing foo.special:
        foo.special = dict(key3=10, key1=80)
        special = foo.special
        special  # --> dict(key3=10, key1=80)
        del special['key1']
        special  # --> dict(key3=10)
        foo.special  # --> dict(key3=10, key1=5)  # key1 was missing, so its setdefault value is used.
        special  # --> dict(key3=10, key1=5)  # a new dict isn't created; foo.special is special.
    '''
    # [TODO] there's probably a cleaner way to achieve this behavior, e.g. make a subclass of dict.
    def apply_key_aliases(self, dict_):
        '''returns dict_ after applying all key_aliases to it.'''
        if isinstance(key_aliases, str):
            aliases = getattr(self, key_aliases)
        else:
            aliases = key_aliases
        for key_alias in aliases:
            with contextlib.suppress(KeyError):
                getattr(self, key_alias)
    def getter(self):
        try:
            result = getattr(self, internal_name)
        except AttributeError:
            result = dict()
            setattr(self, internal_name, result)
        apply_key_aliases(self, result)
        return result
    def setter(self, value):
        setattr(self, internal_name, value)
        apply_key_aliases(self, value)
    def deleter(self):
        delattr(self, internal_name)
    if doc is None:
        doc = f'dict with defaults determined by self.{key_aliases}'
    return property(getter, setter, deleter, doc=doc)



"""
File Purpose: making eppic.i file based on an EppicInstabilityCalculator.
E.g., "make eppic.i file using safe simulation size values, given these physical parameters"
"""
import os
import re

import xarray as xr

from ..eppic_calculator import EppicCalculator
from ....dimensions import XYZ
from ....errors import DimensionalityError, FileContentsMissingError
from ....quantities import QuantityLoader
from ....tools import (
    UNSET,
    is_integer,
    code_snapshot_info,
    xarray_nondim_coords,
)
from ....defaults import DEFAULTS


''' --------------------- Helper functions --------------------- '''

def eppici_dict(ds, *, item=UNSET):
    '''return a dict to use for making an eppic.i file.
    result is like: {
        paramA: valueA,
        paramB: valueB,
        ...  # and similarly, for all global inputs
        0: {
            param0a: value0a,
            param0b: value0b,
            ...  # and similarly, for all dist0 values
        },
        1: {...},
        ...,  # and similarly, for all dists
        N: {...}
    }

    ds: Dataset
        dataset with inputs for eppic.i file.
        data vars will be treated as dist-specific if they contain '{N}' (including curly brackets.)
        internally expands data vars with '{N}' into a value for each dist.
        internally expands data vars with '{x}' into a value for each vector component (x,y,z).
        (if data vars with '{N}' or '{x}' don't vary across 'fluid' or 'component',
            assume that var should have the same value for every fluid or component.)
    item: UNSET or bool
        whether to convert all values to items (i.e., not DataArrays anymore).
        UNSET --> True if dataset dims are only 'fluid' and/or 'component', else False.
    '''
    result = {}

    if item is UNSET:
        item = (set(ds.dims) <= {'fluid', 'component'})
    if item and set(ds.dims) > {'fluid', 'component'}:
        ds_extra_dims = set(ds.dims) - {'fluid', 'component'}
        raise DimensionalityError(f'ds with unrecognized dims when item=True: {ds_extra_dims}')
    to_handle = [(k, v) for k, v in ds.items()]
    # expand '{x}' into components (while maintaining order):
    components = XYZ
    new_to_handle = []
    for var, value in to_handle:
        if '{x}' in var:
            for i, x in enumerate(components):
                vali = value.isel(component=i) if 'component' in value.dims else value
                newvar = var.replace('{x}', str(x))
                new_to_handle.append((newvar, vali))
        else:
            new_to_handle.append((var, value))
    to_handle = new_to_handle
    # put stuff without '{N}' into global params:
    new_to_handle = []
    for var, value in to_handle:
        if '{N}' in var:
            new_to_handle.append((var, value))
            continue
        if item:
            value = value.item()
        result[var] = value
    to_handle = new_to_handle
    # expand '{N}' into fluids, and put into dist params:
    fluids = list(ds['fluid'].values)
    for var, value in to_handle:
        assert '{N}' in var, 'expect all vars to have {N} at this point...'
        for i, f in enumerate(fluids):
            vali = value.isel(fluid=i) if 'fluid' in value.dims else value
            newvar = var.replace('{N}', str(i))
            if item:
                vali = vali.item()
            result.setdefault(i, {})[newvar] = vali
    # done!
    return result

def eppici(ds, dst=None, *,
           meta=True, title=None, coords=True, infos=None,
           TAB=DEFAULTS.TAB, overwrite=False):
    r'''return string suitable for use as an entire eppic.i file.
    (or, return abspath to dst, if wrote to file.)

    ds: Dataset
        dataset with inputs to eppic.i file.
        Must be suitable for being passed to eppici_dict.
        E.g., can use result of EppicInstabilityCalculator('eppici_all').
    dst: None or str
        if provided, write result to the indicated file,
            os.makedirs(exist_ok=True) as needed.
            and return abspath to dst, instead of the usual eppici() result.
    meta: bool
        whether to include metadata near the beginning, as comments.
        E.g., includes datetime when the str was created.
    title: None or str
        if provided, put title=title near the beginning.
    coords: bool, dict, DataArray, or Dataset
        coord info to put after title,
            putting '_coord_' before each key name (to ensure no overlap with eppic inputs.)
            Notes:
                - each coord should probably be a scalar.
                - if any coord has ndim and ndim==0, will be converted to item().
                - string coord will be string (wrapped by single quotes) in eppic.i file.
        True --> use ds.
        DataArray or Dataset --> use coords.nondim_coords(scalars_only=True, item=True).
        Example:
            coords={'kappae': 10.0, 'ngrid_index': 3, 'nice': np.array(7)}
            --> '_coord_kappae = 10.0\n_coord_ngrid_index = 3\n_coord_nice = 7'
    infos: None or dict
        if provided, after coords put this info,
            putting '_info_' before each key name (to ensure no overlap with eppic inputs.)
            Notes:
                - each value should probably be a scalar.
                - if any value has ndim and ndim==0, will be converted to item().
                - string coord will be string (wrapped by single quotes) in eppic.i file.
    TAB: str
        start parameter-defining lines with TAB, to make result prettier.
        (comment lines will not start with tab)
    overwrite: bool
        if dst provided, tells whether to overwrite existing files.
        default False --> by default, never overwrite files.
    '''
    edict = eppici_dict(ds, item=True)  # must use item=True if writing to string!
    lines = []
    if meta:
        # PlasmaCalcs meta
        mstr = '; Created by PlasmaCalcs'
        info = code_snapshot_info()
        if 'pc__version__' in info:
            mstr = mstr + f" version {info['pc__version__']}"
        if 'pc__commit_hash' in info:
            mstr = mstr + f" (commit {info['pc__commit_hash']})"
        if 'datetime' in info:
            mstr = mstr + f"\n; {TAB}at time={info['datetime']}"
        lines.append(mstr)
        # instructions for where to learn meanings of parameters
        lines.append('; For parameter meanings, see src/eppic.h, from gitlab.com/oppenheim/eppic')
        # [TODO] other meta, e.g. values of safety factors,
        #  or other convenient parameters to know at a glance, e.g. temperatures?
    if title is not None:
        lines.append('')
        lines.append(f"title = '{title}'")
    if coords == True:
        coords = ds
    if isinstance(coords, (xr.DataArray, xr.Dataset)):
        coords = xarray_nondim_coords(coords, scalars_only=True, item=True)
    if coords != False:
        lines.append('')
        lines.append('; Coords (ignored by eppic. For bookkeeping purposes.)')
        for key, value in coords.items():
            try:
                value = f'{value:.4g}'
            except (ValueError, TypeError):
                value = f"'{value}'" if isinstance(value, str) else str(value)
            lines.append(f"{TAB}_coord_{key} = {value}")
    if infos is not None:
        lines.append('')
        lines.append('; Infos (ignored by eppic. For bookkeeping purposes.)')
        for key, value in infos.items():
            if hasattr(value, 'ndim') and value.ndim == 0: value = value.item()
            try:
                value = f'{value:.4g}'
            except (ValueError, TypeError):
                value = f"'{value}'" if isinstance(value, str) else str(value)
            lines.append(f"{TAB}_info_{key} = {value}")
    # global parameters:
    lines.append('')
    lines.append('')
    lines.append('; Global parameters')
    for var, value in list(edict.items()):
        if isinstance(var, int):  # dist-info skipped for now
            continue
        if hasattr(value, 'ndim') and value.ndim == 0: value = value.item()
        if not is_integer(value):  # never convert ints to sci notation for eppic.i
            try:
                value = f'{value:.7g}'  # 7 sigfigs should be plenty.
            except (ValueError, TypeError):
                pass  # it's fine, just use default string conversion then.
        lines.append(f'{TAB}{var} = {value}')
        del edict[var]
    # fluid-specific parameters:
    lines.append('')
    lines.append('')
    lines.append('; Fluid parameters')
    lines.append(f'{TAB}ndist = {len(ds["fluid"])}')
    for i, fdict in edict.items():
        lines.append('')
        lines.append('')
        fname = str(ds['fluid'].isel(fluid=i).item())
        lines.append(f'; ----- {fname} -----')
        lines.append(f'{TAB}dist = {i}')
        for var, value in fdict.items():
            if hasattr(value, 'ndim') and value.ndim == 0: value = value.item()
            if not is_integer(value):  # never convert ints to sci notation for eppic.i
                try:
                    value = f'{value:.7g}'  # 7 sigfigs should be plenty.
                except (ValueError, TypeError):
                    pass  # it's fine, just use default string conversion then.
            lines.append(f'{TAB}{var} = {value}')
    lines.append('')
    result = '\n'.join(lines)
    # write to file, if relevant.
    if dst is None:
        return result
    else:
        if os.path.exists(dst) and not overwrite:
            raise FileExistsError(f'{dst!r}, when overwrite=False')
        os.makedirs(os.path.dirname(dst), exist_ok=True)
        with open(dst, 'w' if overwrite else 'x') as f:
            f.write(result)
        return os.path.abspath(dst)


''' --------------------- EppiciMaker --------------------- '''

class EppiciMaker(QuantityLoader):
    '''makes eppic.i input file based on an EppicInstabilityCalculator.
    E.g., "make eppic.i file using safe simulation size values, given these physical parameters".
    For details on inputs, see EppicGlobInputsLoader and EppicDistInputsLoader.
    '''

    # # # RENAMING DIST & GLOB INPUTS TO EPPICI NAMES # # #

    @known_var(deps=['dist_inputs'])
    def get_eppici_dist(self):
        '''Dataset of dist-related input values which go to eppic.i, using eppic.i datavar names.

        Similar to self('dist_inputs') however vars have been renamed to eppic.i names,
            with {x} indicating "replace with component (x, y, or z)",
            and {N} indicating "replace with fluid index".
        Still using PlasmaCalcs dimensions.
            e.g. var='pv{x}min{N}' (keeping curly braces!) varies across 'component' and 'fluid'.
        
        Result keys determined by renaming dist_inputs keys via self.DIST_INPUTS.
            (dict-valued keys there are expanded here into multiple keys)

        if self.dspace_mode doesn't start with 'safe_', use 'safe_{mode}' instead.
            (this affects, e.g., safe_pow2_subcycle)
        '''
        result = self('dist_inputs')
        return self._handle_renaming_ds(result, self.DIST_INPUTS)

    @known_var(deps=['glob_inputs'])
    def get_eppici_glob(self):
        '''Dataset of global input values which go to eppic.i, using eppic.i datavar names.
        Similar to self('glob_inputs') however vars have been renamed to eppic.i names,
            with {x} indicating "replace with component (x, y, or z)".
        Still using PlasmaCalcs dimensions.
            e.g. var='E{x}0_external' (keeping curly braces!) varies across 'component'.
        internally uses self.dspace_mode = self.safe_dspace_mode.
        Result keys are determined by renaming glob_inputs keys via self.GLOB_INPUTS.
            (dict-valued keys there are expanded here into multiple keys)
        '''
        result = self('glob_inputs')
        return self._handle_renaming_ds(result, self.GLOB_INPUTS)

    @staticmethod
    def _renaming_instructions(renaming):
        '''flatten dict like GLOB_INPUTS into a single dict, for renaming.
        Result order is in order used by renaming, with dict values expanded in-place.
        '''
        result = {}
        for var, name in list(renaming.items()):
            if isinstance(name, dict):
                for subvar, subname in name.items():
                    result[subvar] = subname
            else:
                result[var] = name
        return result

    def _handle_renaming_ds(self, ds, renaming):
        '''handle renaming the vars in ds according to a dict like GLOB_INPUTS.
        Result order is in order implied by renaming, when dict values expanded in-place.
        '''
        renaming = self._renaming_instructions(renaming)
        result = ds.rename(renaming)
        # style: sort data vars by order implied by renaming.
        newnames = list(renaming.values())
        assert set(result.data_vars) == set(newnames), 'unexpected var in result...'
        return result[newnames]


    # # # COMBINING DIST & GLOB INPUTS # # #

    @known_var(deps=['glob_inputs', 'dist_inputs'])
    def get_all_inputs(self):
        '''Dataset of all input values which go to eppic.i, here using PlasmaCalcs datavar names.

        Named & dimensioned here with PlasmaCalcs conventions,
            e.g. uses 'E_un0' which varies across 'component' dim,
            instead of labeling Ex0_external, Ey0_external, Ez0_external like in eppic.i.
        Result has all keys from dist_inputs and glob_inputs;
            see self.help('dist_inputs') and self.help('glob_inputs') for more details.

        if self.dspace_mode doesn't start with 'safe_', use 'safe_{mode}' instead.
            (this affects, e.g., safe_pow2_subcycle)
        '''
        return self(['glob_inputs', 'dist_inputs'], dspace_mode=self.safe_dspace_mode)

    @known_var(deps=['eppici_glob', 'eppici_dist'])
    def get_eppici_all(self):
        '''Dataset of all input values which go to eppic.i, using eppic.i datavar names.

        Similar to self('all_inputs') however vars have been renamed to eppic.i names,
            with {x} indicating "replace with component (x, y, or z)",
            and {N} indicating "replace with fluid index".
        Still using PlasmaCalcs dimensions.
            e.g. var='pv{x}min{N}' (keeping curly braces!) varies across 'component' and 'fluid'.
        
        Result keys are determined by renaming all_inputs keys via self.DIST_INPUTS and self.GLOB_INPUTS.
            (dict-valued keys there are expanded here into multiple keys)

        if self.dspace_mode doesn't start with 'safe_', use 'safe_{mode}' instead.
            (this affects, e.g., safe_pow2_subcycle)
        '''
        return self(['eppici_glob', 'eppici_dist'])


    # # # MAKING EPPIC.I FILE # # #

    def eppici_dict(self, ds=None, *, item=UNSET):
        '''return a dict to use for making an eppic.i file.
        result has global params as keys,
        also fluid indexes (e.g. 0, 1, 2) as keys, with dict values telling fluid params.

        ds: None or Dataset
            dataset with inputs for eppic.i file.
            None --> self('eppici_all').
        item: UNSET or bool
            whether to convert all values to items (i.e., not DataArrays anymore).
            UNSET --> True if dataset dims are only 'fluid' and/or 'component', else False.
        '''
        if ds is None:
            ds = self('eppici_all')
        return eppici_dict(ds, item=item)

    def eppici(self, dst=None, *, ds=None, title=None, coords=True,
               infos=None, safety_infos=True, safe_sim_size_infos=True,
               makeslurm='attempt', **kw_eppici):
        '''return string for new eppic.i file contents, or write to dst and return abspath.
        uses 'safe' values from self to decide params: ds=self('eppici_all') if not provided.
        
        dst: None or str
            str --> filepath for where to write result; os.makedirs(exist_ok=True) as needed.
                    And, this function returns abspath to dst.
            None --> do not write to any file.
                    And, this function returns string for eppic.i contents.
        ds: None or Dataset
            dataset with inputs to eppic.i file. Passed to eppici_dict.
            None --> uses self('eppici_all').
        title: None or str
            if provided, put title=title near the beginning.
        coords: bool, dict, DataArray, or Dataset
            coord info to put after title, replacing keys with '_coord_{key}'.
            True --> infer coords from scalar nondim_coords of ds.
        infos: None or dict
            additional infos to put after coords, replacing keys with '_info_{key}'.
        safety_infos: bool
            whether to include self('safety_details') values, replacing keys with '_info_{key}'.
        safe_sim_size_infos: bool
            whether to include self('safe_sim_size') values, replacing keys with '_info_{key}'.
        makeslurm: 'attempt', True, False, or dict
            whether to call self.makeslurm() after making eppic.i file (if dst provided).
            'attempt' --> try it, but skip if can't find slurm template file.
            True --> call self.makeslurm(). If it fails, crash.
            False --> don't self.makeslurm().
            dict --> dict of kwargs to pass to self.makeslurm().
        overwrite: bool
            if dst provided, tells whether to overwrite existing files.
            default False --> by default, never overwrite files.

        additional kwargs go to the eppici helper function,
            from hookups/eppic/eppic_instability_calculator/eppici_maker.py
        '''
        if ds is None:
            ds = self('eppici_all')
        if safety_infos or safe_sim_size_infos:
            infos = {} if infos is None else {**infos}
        if safety_infos:
            infos.update(self('safety_details'))
        if safe_sim_size_infos:
            infos.update(self('safe_sim_size'))
        result = eppici(ds, dst=dst, title=title, coords=coords, infos=infos, **kw_eppici)
        if dst is None or makeslurm == False:
            return result
        else:
            eppici_path = result
            if makeslurm == True:
                makeslurm = {}
            if makeslurm == 'attempt':
                try:
                    self.makeslurm(eppici_path)
                except FileNotFoundError:
                    pass
            else:
                self.makeslurm(eppici_path, **makeslurm)
            return result

    def _slurm_template_path(self, eppici, template=None):
        '''return path to template slurm file. If not provided, infer it:
            (1) ec = EppicCalculator.from_here(eppici)  # eppici is filepath
            (2) look in ec.notes_dirname for slurm file with appropriate basename
                ("appropriate" means .i --> .slurm. E.g. 'eppic_0.i' --> 'eppic_0.slurm')
            (3) look in os.path.dirname(ec.notes_dirname) for slurm file like that^
            (4) give up, crash with FileNotFoundError
        If provided, check to make sure it exists, then return abspath to it.
        '''
        if template is not None:
            template = os.path.abspath(template)
            if not os.path.exists(template):
                raise FileNotFoundError(f'slurm template not found: {template!r}')
            return template
        # else
        eppici = os.path.abspath(eppici)  # <-- ensures result will be abspath
        ec = EppicCalculator.from_here(eppici, u_t=1)  # u_t=1 prevents pic ambiguous unit crash.
        ec_notes = ec.notes_dirname
        basename = os.path.splitext(os.path.basename(eppici))[0] + '.slurm'
        template2 = os.path.join(ec_notes, basename)
        if os.path.exists(template2):
            return template2
        template3 = os.path.join(os.path.dirname(ec_notes), basename)
        if os.path.exists(template3):
            return template3
        errmsg = f'slurm template for{eppici!r}.\nChecked {template2!r}\nand {template3!r}'
        raise FileNotFoundError(errmsg)

    def makeslurm(self, eppici, template=None, *, dst=UNSET):
        '''write a slurm file using details from eppic.i file as input.

        Expects template file to have lines like:
            #SBATCH -N 8     # possibly with comments too.
            #SBATCH --tasks-per-node=56
            #SBATCH -t 01:00:00
        which specify number of nodes, tasks per node, and wall clock time requested.
        These will be replaced by params from eppici file:
            _info_n_nodes,
            _info_tasks_per_node,
            _info_safe_runtime_HMS.
        ([TODO] more flexible slurm options, e.g. specifying n_processors instead of n_nodes.)

        eppici: str
            path to eppic.i file.
        template: None or str
            path to template slurm file.
            None --> infer from eppici path. Check in default notes_dirname associated with eppici,
                    for file with similar basename but .slurm (e.g. 'eppic_0.i' --> 'eppic_0.slurm').
                    If that doesn't exist, check in dirname of notes_dirname.
                    If that also doesn't exist, crash.
        dst: UNSET, None, or str
            name for the resulting slurm file.
            UNSET --> same path as eppici path, but using .slurm extension instead.
            None --> return str of slurm contents instead of making slurmfile.
            str --> write to that path, os.makedirs(exist_ok=True) as needed.

        returns abspath to created slurmfile. (unless dst=None; see above for details.)
        '''
        if dst is UNSET:
            dst = os.path.splitext(eppici)[0] + '.slurm'
        template = self._slurm_template_path(eppici, template=template)
        ec = EppicCalculator.from_here(eppici, u_t=1)  # u_t=1 prevents pic ambiguous unit crash.
        slurmkeys = ['_info_n_nodes', '_info_tasks_per_node', '_info_safe_runtime_HMS']
        _missing = [k for k in slurmkeys if k not in ec.input_deck.keys()]
        if len(_missing) > 0:
            errmsg = f'expected all keys {slurmkeys} in eppic.i file; missing {_missing}'
            raise FileContentsMissingError(errmsg)
        slurmvals = {k: ec.input_deck[k] for k in slurmkeys}
        with open(template, 'r') as f:
            slurmlines = f.readlines()
        to_match = {
            '_info_n_nodes': r'(#SBATCH)(\s+)(-N)(\s+)(\d+)(\s*)(#.*)?',
            '_info_tasks_per_node': r'(#SBATCH)(\s+)(--tasks-per-node=)()(\d+)(\s*)(#.*)?',
            '_info_safe_runtime_HMS': r'(#SBATCH)(\s+)(-t)(\s+)(\d+:\d+:\d+)(\s*)(#.*)?',
        }
        for i, line in enumerate(slurmlines):
            for key, pattern in to_match.items():
                match = re.match(pattern, line)
                if match is not None:
                    sbatch, space0, option, space1, value, space2, comment = match.groups()
                    slurmlines[i] = f'{sbatch}{space0}{option}{space1}{slurmvals[key]}{space2}{comment}\n'
                    to_match.pop(key)
                    break
        if len(to_match) > 0:
            errmsg = f'template slurm file missing lines like: {list(to_match.values())}'
            raise FileContentsMissingError(errmsg)
        result = ''.join(slurmlines)
        if dst is None:
            return result
        else:
            with open(dst, 'w') as f:
                f.write(result)
            return os.path.abspath(dst)



"""
File Purpose: misc. tools for reading directly from bifrost files
"""
import os
import re

import numpy as np

from ...errors import (
    InputError,
    FileAmbiguityError, FileContentsError, FileContentsConflictError,
)
from ...tools import (
    read_idl_params_file, updated_idl_params_file,
    alias_child, alias_key_of, weakref_property_simple,
    UNSET,
    ProgressUpdater,
)


''' --------------------- bifrost snapname_NNN.idl files --------------------- '''

def read_bifrost_snap_idl(filename, *, eval=True, strip_strs=True):
    '''Parse Bifrost snapname_NNN.idl file into a dictionary.

    filename: string
        file to read from.
    eval: bool, default True
        whether to attempt to evaluate the values,
        using ast.literal_eval (safer but less flexible than eval).
        False --> values will remain as strings.
        True --> try to evaluate values but use strings if evaluation fails.
            Also convert '.true.' and '.false.' to True and False, case-insensitively.
    strip_strs: bool, default True
        whether to strip whitespace from all string values in result, after eval.

    File formatting notes:
        - semicolons (;) are used for comments. (idl format)
        - ignore blank lines & lines that don't assign a variable (missing '=')
        - ignores all leading & trailing whitespace in vars & values.
        - might use '.true.' and '.false.' for booleans.
    '''
    result = read_idl_params_file(filename, eval=eval)
    if eval:
        # handle .true. and .false. conversions below. Other evals have already been handled^
        for key, value in result.items():
            if isinstance(value, str):
                lowered = value.lower()
                if lowered == '.true.':
                    result[key] = True
                elif lowered == '.false.':
                    result[key] = False
        # strip whitespace from string values in result
        if strip_strs:
            for key, value in result.items():
                if isinstance(value, str):
                    result[key] = value.strip()
    return result

def bifrost_snap_idl_files(snapname, *, dir=os.curdir, abspath=False):
    '''return list of all bifrost snapname_NNN.idl files in directory.
    Sorts by snap number.

    snapname: str. match snapname_NNN.idl. (NNN can be any integer, doesn't need to be 3 digits.)
    abspath: bool. whether to return absolute paths or just the file basenames within directory.
    '''
    pattern = rf'{snapname}_([0-9]+)[.]idl'
    result = []
    for f in os.listdir(dir):
        match = re.fullmatch(pattern, f)
        if match is not None:
            snap_number = int(match.group(1))
            result.append((snap_number, f))
    result.sort()
    result = [f for (n, f) in result]
    if abspath:
        absdir = os.path.abspath(dir)
        result = [os.path.join(absdir, f) for f in result]
    return result

def bifrost_infer_snapname_here(dir=os.curdir):
    '''infer snapname based on files in directory, if possible.
    For files like snapname_NNN.idl, if all have same snapname, return it.
    If no such files, raise FileNotFoundError; if multiple implied snapnames, raise FileAmbiguityError.
    NNN can be any integer, doesn't need to be 3 digits.
    '''
    pattern = re.compile(r'(.+)_[0-9]+[.]idl')
    snapname = None
    for f in os.listdir(dir):
        match = pattern.fullmatch(f)
        if match is not None:
            if snapname is None:
                snapname = match.group(1)
            elif snapname != match.group(1):
                raise FileAmbiguityError(f'found different snapnames: {snapname!r}, {match.group(1)!r}')
    if snapname is None:
        raise FileNotFoundError(f'no files like "snapname_NNN.idl" found in directory: {dir!r}')
    return snapname


''' --------------------- bifrost mesh files --------------------- '''

def read_bifrost_meshfile(meshfile):
    '''returns dict of mesh coords from a Bifrost mesh file.

    Mesh file format looks like:
        x size (int)
        x coords
        x "down" coords
        dx values when taking "up derivative"
        dx values when taking "down derivative"
        then similar for y and z.

    The "down" and "up" refer to interpolation / staggering.

    result will have keys:
        x_size: int, number of points in x
        x: x coords list (as numpy array)
        x_down: x "down" coords
        dxup: dx when taking "up derivative"
        dxdn: dx when taking "down derivative"
    '''
    meshfile = os.path.abspath(meshfile)  # <-- makes error messages more verbose, if crash later.
    with open(meshfile, 'r') as f:
        lines = f.readlines()
    if len(lines) != 5 * len(('x', 'y', 'z')):
        raise FileContentsError(f'expected 5 lines per axis, got nlines={len(lines)}')
    result = {}
    x_to_lines = {'x': lines[:5], 'y': lines[5:10], 'z': lines[10:15]}
    for x, xlines in x_to_lines.items():
        result[f'{x}_size'] = int(xlines[0])
        result[x]           = np.array([float(s) for s in xlines[1].split()])
        result[f'{x}_down'] = np.array([float(s) for s in xlines[2].split()])
        result[f'{x}_ddup'] = np.array([float(s) for s in xlines[3].split()])
        result[f'{x}_dddn'] = np.array([float(s) for s in xlines[4].split()])
        # sanity checks:
        for key in [x, f'{x}_down', f'{x}_ddup', f'{x}_dddn']:
            if len(result[key]) != result[f'{x}_size']:
                raise FileContentsConflictError(f'length of {key!r} does not match {x}_size')
    return result

def slice_bifrost_meshfile(meshfile, slices, *, dst=None):
    '''read meshfile contents, slice by slices, output to dst (return as str if dst=None)
    meshfile: str, path to mesh file
    slices: dict of {x: indexer} for any x from ('x', 'y', 'z').
        currently, implementation requires all provided indexers to be slice objects.
    dst: None or str.
        None --> return output as a string.
        str --> write output to file with this name; return abspath to dst.
                (never overwrites existing output; dst must be a file that doesn't exist yet.)
    '''
    # read meshfile. (to keep all as strings, we avoid using read_bifrost_meshfile directlY)
    meshfile = os.path.abspath(meshfile)
    with open(meshfile, 'r') as f:
        lines = f.readlines()
    if len(lines) != 5 * len(('x', 'y', 'z')):
        raise FileContentsError(f'expected 5 lines per axis, got nlines={len(lines)}')
    lines = [line.rstrip('\n') for line in lines]
    result = []
    x_to_lines = {'x': lines[:5], 'y': lines[5:10], 'z': lines[10:15]}
    for x in ('x', 'y', 'z'):
        xlines = x_to_lines[x]
        if x in slices:
            indexer = slices[x]
            if not isinstance(indexer, slice):
                raise NotImplementedError(f'expected slices[{x!r}] to be a slice; got {indexer!r}')
            new_arr_lines = []
            new_size = None
            for line in xlines[1:]:  # x, x_down, x_ddup, x_dddn'
                splitted = line.split()
                new_vals = splitted[indexer]
                new_line = '   ' + '   '.join(new_vals)  # spacing matches default mesh formatting for arr lines.
                new_arr_lines.append(new_line)
                if new_size is None:
                    new_size = len(new_vals)
                else:
                    assert new_size == len(new_vals), 'expect all arr_lines to have same length'
            result.append(f'{new_size:>12d}')  # >12d matches default mesh formatting for size lines.
            result.extend(new_arr_lines)
        else:  # x is not being sliced; don't change anything!
            result.extend(xlines)
    result = '\n'.join(result)
    # output result
    if dst is None:
        return result
    else:
        with open(dst, 'x') as f:   # mode='x' --> crash if dst exists; never touch existing files.
            f.write(result)
        return os.path.abspath(dst)


''' --------------------- BifrostVarPathsManager --------------------- '''

class BifrostVarPathsManager():
    '''manages filepaths (as abspaths) and readable vars for a BifrostSnap.
    self.kind2path = {kind: path}
    self.kind2vars = {kind: [list of readable vars]}
    self.var2kind = {var: kind}
    self.var2path = {var: path}
    self.path2kind = {path: kind}
    self.path2vars = {path: [list of readable vars]}
    self.var2index = {var: index of var in its path's list of vars}
    
    self.kinds: tuple of kinds with any vars in self.
    self.vars: tuple of all vars in self.
    self.paths: tuple of all paths with any vars in self.

    kinds are: 'snap', 'aux', 'hion', 'helium', 'ooe'.
    if kind has no vars, do not include it in results.

    snap: BifrostSnap
    bcalc: BifrostCalculator
    '''
    KINDS = ('snap', 'aux', 'hion', 'helium', 'ooe')
    KIND2FILEBASE = {
        'snap': '{snapname}_{NNN}.snap',
        'aux': '{snapname}_{NNN}.aux',
        'hion': '{snapname}.hion_{NNN}.snap',
        'helium': '{snapname}.helium_{NNN}.snap',
    }

    def __init__(self, snap, bcalc):
        self.snap = snap
        self.bcalc = bcalc
        self.init_all()

    snap = weakref_property_simple('_snap')  # weakref --> snap caching paths manager would be fine.
    bcalc = weakref_property_simple('_bcalc')  # weakref --> bcalc caching paths manager would be fine.
    params = alias_child('snap', 'params')
    snapname = alias_key_of('params', 'snapname')
    NNN = property(lambda self: self.snap.file_s(self.bcalc),
            doc='''(str) the NNN part of the snapname_NNN.idl filename.''')
    snapdir = alias_child('bcalc', 'snapdir')

    def snappath(self, filename):
        '''returns os.path.join(self.snapdir, filename)'''
        return os.path.join(self.snapdir, filename)

    def kind2filebase(self, kind):
        '''returns file basename for this kind.
        Equivalent: self.KIND2FILEBASE[kind].format(snapname=self.snapname, NNN=self.NNN)
        '''
        return self.KIND2FILEBASE[kind].format(snapname=self.snapname, NNN=self.NNN)

    def init_all(self):
        '''init all KINDS in self.'''
        self.kind2path = dict()
        self.kind2vars = dict()
        self.var2kind = dict()
        self.var2path = dict()
        self.path2kind = dict()
        self.path2vars = dict()
        self.var2index = dict()
        self.init_snap_kind()
        self.init_aux_kind()
        self.init_hion_kind()
        self.init_helium_kind()
        self.init_ooe_kind()
        self.kinds = tuple(self.kind2vars.keys())
        self.vars = tuple(self.var2kind.keys())
        self.paths = tuple(self.path2vars.keys())

    def _init_kind_vars_path(self, kind, vars, path):
        '''updates self with corresponding kind, vars, and path.'''
        self.kind2vars[kind] = vars
        self.kind2path[kind] = path
        for var in vars:
            if var in self.var2kind:  # var not unique... crash!
                errmsg = f'{type(self).__name__} with multiple vars with same name: {var!r}'
                raise LoadingNotImplementedError(errmsg)
            self.var2kind[var] = kind
            self.var2path[var] = path
        self.path2kind[path] = kind
        self.path2vars[path] = vars
        self.var2index.update({var: i for i, var in enumerate(vars)})

    def init_snap_kind(self):
        '''vars stored in snapname_NNN.snap file.'''
        path = self.snappath(self.kind2filebase('snap'))
        if self.params.get('do_mhd', False):
            vars = ('r', 'px', 'py', 'pz', 'e', 'bx', 'by', 'bz')
        else:
            vars = ('r', 'px', 'py', 'pz', 'e')
        self._init_kind_vars_path('snap', vars, path)

    def init_aux_kind(self):
        '''vars stored in snapname_NNN.aux file.'''
        path = self.snappath(self.kind2filebase('aux'))
        vars = tuple(self.params.get('aux', '').split())
        if len(vars) > 0:
            self._init_kind_vars_path('aux', vars, path)

    def init_hion_kind(self):
        '''vars stored in snapname.hion_NNN.snap file.'''
        path = self.snappath(self.kind2filebase('hion'))
        if self.params.get('do_hion', 0) > 0:
            vars = ('hionne', 'hiontg', 'n1', 'n2', 'n3', 'n4', 'n5', 'n6', 'nh2')
            self._init_kind_vars_path('hion', vars, path)

    def init_helium_kind(self):
        '''vars stored in snapname.helium_NNN.snap file.'''
        path = self.snappath(self.kind2filebase('helium'))
        if self.params.get('do_helium', 0) > 0:
            vars = ('nhe1', 'nhe2', 'nhe3')
            self._init_kind_vars_path('helium', vars, path)

    def init_ooe_kind(self):
        '''out of equilibrium vars.'''
        if self.params.get('do_out_of_eq', 0) > 0:
            pass  # fail silently; not implemented yet.
            # (doesn't give WRONG answer, just gives no answer, so it's okay to fail silently)
            #raise NotImplementedError('loading ooe vars. Got do_out_of_eq > 0')

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}(vars={self.vars})'

    def help(self):
        '''print docstring of self...'''
        print(type(self).__doc__)


class BifrostScrVarPathsManager(BifrostVarPathsManager):
    '''manages filepaths (as abspaths) and readable vars for a SCR_SNAP
    self.kind2path = {kind: path}
    self.kind2vars = {kind: [list of readable vars]}
    self.var2kind = {var: kind}
    self.var2path = {var: path}
    self.path2kind = {path: kind}
    self.path2vars = {path: [list of readable vars]}
    self.var2index = {var: index of var in its path's list of vars}
    
    self.kinds: tuple of kinds with any vars in self.
    self.vars: tuple of all vars in self.
    self.paths: tuple of all paths with any vars in self.

    kinds are: 'snap', 'aux', 'hion', 'helium', 'ooe'.
    if kind has no vars, do not include it in results.

    snap: BifrostSnap
    bcalc: BifrostCalculator
    '''
    KIND2FILEBASE = {
        'snap': '{snapname}.snap.scr',
        'aux': '{snapname}.aux.scr',
        'hion': '{snapname}.hion.snap.scr',
        'helium': '{snapname}.helium.snap.scr',
    }


class BifrostDataCutter(BifrostVarPathsManager):
    '''interface for creating new data files cut from existing data files.
    "cut" refers to indexing by simple slices, with trivial step (1 or None),
        e.g. slice x by slice(500, 1000). But not by slice(500, 1000, 10).

    (NEVER overwrites existing files; if any result files already exist, crash.)

    self.kind2path = {kind: path}
    self.kind2vars = {kind: [list of readable vars]}
    self.var2kind = {var: kind}
    self.var2path = {var: path}
    self.path2kind = {path: kind}
    self.path2vars = {path: [list of readable vars]}
    self.var2index = {var: index of var in its path's list of vars}

    self.kind2newpath = {kind: newpath}
    self.path2newpath = {original path: newpath}
    self.newpath2path = {newpath: original path}
    
    self.kinds: tuple of kinds with any vars in self.
    self.vars: tuple of all vars in self.
    self.paths: tuple of all paths with any vars in self.
    self.newpaths: tuple of all newpaths with any vars in self.

    kinds are: 'snap', 'aux', 'hion', 'helium', 'ooe'.
    if kind has no vars, do not include it in results.

    snap: BifrostSnap
    bcalc: BifrostCalculator
    new_snapname: str or UNSET
        new snapname to use for the cutted data files. UNSET --> '{self.snapname}_cut'
    '''
    def __init__(self, snap, bcalc, *, new_snapname=UNSET):
        super().__init__(snap, bcalc)
        self.new_snapname = f'{self.bcalc.snapname}_cut' if new_snapname is UNSET else new_snapname
        self.init_new_paths()

    def newpath(self, oldpath):
        '''return newpath from oldpath. (both are abspaths.)
        newpath = oldpath, but replace self.bcalc.snapname (within oldpath basename) by new_snapname
        crash if newpath == oldpath.
        '''
        oldbasename = os.path.basename(oldpath)
        newbasename = oldbasename.replace(self.bcalc.snapname, self.new_snapname)
        newpath = os.path.join(os.path.dirname(oldpath), newbasename)
        if newpath == oldpath:
            raise FileExistsError(f'Expect newpath != oldpath, but got both equal to: {newpath!r}')
        return newpath

    def init_new_paths(self):
        '''initialize self.kind2newpath, path2newpath and newpath2path'''
        self.kind2newpath = dict()
        self.path2newpath = dict()
        self.newpath2path = dict()
        for kind, path in self.kind2path.items():
            newpath = self.newpath(path)
            self.kind2newpath[kind] = newpath
            self.path2newpath[path] = newpath
            self.newpath2path[newpath] = path
        self.newpaths = tuple(self.kind2newpath.values())

    # # # SLICES # # #
    @property
    def slices(self):
        '''self.bcalc.slices, but ensuring all are slice objects and have trivial step (1 or None).
        Also ensures at least 1 nontrivial slice exists (i.e. not empty and not all slice(None).)'''
        slices = self.bcalc.slices
        if not all(isinstance(s, slice) for s in slices.values()):
            raise NotImplementedError(f'cut, when non-slice object(s) in slices; got slices={slices!r}')
        if not all(s.step is None or s.step==1 for s in slices.values()):
            raise NotImplementedError(f'cut, with nontrivial step (step != 1 or None); got slices={slices!r}')
        if all(s == slice(None) for s in slices.values()):
            raise InputError(f'no reason to cut; no non-trivial slices exist. Got slices={slices!r}')
        return slices

    def slices_tuple(self):
        '''returns tuple of slices to apply to array with (x, y, z) as first 3 dims.
        determined by self.slices.get(x, slice(None)) for x in 'x', 'y', 'z'.
        '''
        slices = self.slices
        return tuple(slices.get(x, slice(None)) for x in ('x', 'y', 'z'))

    # # # CUT KIND # # #
    def cut_kind(self, kind):
        '''cut data for this kind, writing results to self.kind2newpath[kind].
        see self.kinds for options.
        returns self.kind2newpath[kind], which is an abspath to the new path.

        (NEVER overwrites existing data; if the newpath exists already, crash.)

        amount to cut is determined by self.bcalc.slices.
        size of arrays within memmap data files is determined by bcalc.data_array_shape.
        '''
        # preprocessing
        oldpath = self.kind2path[kind]
        newpath = self.kind2newpath[kind]
        if os.path.exists(newpath):
            raise FileExistsError(f'newpath already exists: {newpath!r}')
        array_shape = self.bcalc.data_array_shape
        _vars_here = self.kind2vars[kind]
        n_arrays_here = len(_vars_here)
        full_file_shape = (*array_shape, n_arrays_here)   # (nx, ny, nz, n_arrays_here)
        slices_tuple = self.slices_tuple()
        kw_memmap = dict(dtype=self.bcalc.data_array_dtype,
                         order=self.bcalc.data_array_order)
        # reading data
        data = np.memmap(oldpath, offset=0, mode='r',  # read-only; never alters existing files!
                         shape=full_file_shape, **kw_memmap)
        # slicing data
        newdata = data[slices_tuple]
        # writing data
        mapped = np.memmap(newpath, offset=0, mode='w+',
                           shape=newdata.shape, **kw_memmap)
        mapped[:] = newdata[:]
        mapped.flush()
        return newpath

    # # # CUT MESHFILE # # #
    def _new_meshfile_path(self):
        '''abspath to new meshfile. Equivalent: self.newpath(self.bcalc.meshfile).'''
        return self.newpath(self.bcalc.meshfile)

    def cut_meshfile(self):
        '''cut meshfile, writing results to self.newpath(self.bcalc.meshfile).
        amount to cut is determined by self.bcalc.slices.

        "cutting the meshfile" means apply slice_bifrost_meshfile()
        '''
        oldpath = self.bcalc.meshfile
        newpath = self._new_meshfile_path()
        return slice_bifrost_meshfile(oldpath, self.slices, dst=newpath)

    # # # CUT IDL FILE # # #
    def _idl_file_path(self):
        '''path to idl file for self.snap. Equivalent: self.bcalc.snap_filepath(self.snap).'''
        return self.bcalc.snap_filepath(self.snap)

    def _new_idl_file_path(self):
        '''abspath to new idl file. Equivalent: self.newpath(self.bcalc.snap_filepath(self.snap)).'''
        return self.newpath(self._idl_file_path())

    def _data_array_shape_after_cut(self):
        '''returns shape each data array will be after cutting (using self.slices).'''
        slices = self.slices  # self.slices property guarantees each value is a slice with trivial step.
        shape = list(self.bcalc.data_array_shape)
        assert len(shape)==3, 'not implemented in non-3D case...'
        for i, x in enumerate(('x', 'y', 'z')):
            if x in slices:
                indexer = slices[x]
                new_idx = range(shape[i])[indexer]
                new_len = len(new_idx)
                shape[i] = new_len
        return tuple(shape)

    def cut_idl_file(self):
        '''cut idl_file, writing results to self.newpath(self.bcalc.idlfile).
        the idlfile is the snapname_NNN.idl file; see self.bcalc.snap_filepath().
        amount to cut is determined by self.bcalc.slices.

        "cutting the idlfile" means replacing the relevant parameters in the idlfile:
            'mx', 'my', 'mz' --> new sizes after cutting
            'snapname' --> new snapname
            'meshfile' --> new meshfile name
        '''
        new_shape = self._data_array_shape_after_cut()
        params = dict()
        params['mx'] = new_shape[0]
        params['my'] = new_shape[1]
        params['mz'] = new_shape[2]
        new_snapname = self.new_snapname
        params['snapname'] = f'"{new_snapname}"'
        new_meshfile = os.path.basename(self.newpath(self.bcalc.meshfile))
        params['meshfile'] = f'"{new_meshfile}"'
        return updated_idl_params_file(self._idl_file_path(), params, dst=self._new_idl_file_path())

    # # # CUT ALL # # #
    def new_file_paths(self):
        '''list of abspaths to all new files which self.cut() will create.
        These are self.newpaths, and the new path for the meshfile & idl file.
        '''
        return list(self.newpaths) + [self._new_meshfile_path(), self._new_idl_file_path()]

    def _check_output_files_dont_exist(self):
        '''ensure that all output files do not exist yet.'''
        for path in self.new_file_paths():
            if os.path.exists(path):
                raise FileExistsError(f'output file already exists: {path!r}')

    def cut(self, *, verbose=True):
        '''cut all data & supporting files, writing results to new path based on self.new_snapname.
        (never edits any existing files, only makes new ones.)

        returns list of all new file paths (i.e. self.new_file_paths())

        verbose: bool
            whether to print progress updates during cutting.
        '''
        updater = ProgressUpdater(print_freq=0 if verbose else -1)
        self._check_output_files_dont_exist()
        updater.print(f'cutting idl file')
        self.cut_idl_file()
        updater.print(f'cutting meshfile')
        self.cut_meshfile()
        # ^ do those first in case of crash; they are fast & inexpensive.
        # v next do the data files.
        for i, kind in enumerate(self.kinds):
            updater.print(f'cutting {kind} data (this is kind {i+1} of {len(self.kinds)})')
            self.cut_kind(kind)
        return self.new_file_paths()



"""Consistent interface for plasma calculations from any inputs.

[TODO] more details about PlasmaCalcs here.
"""
#[TODO] explicit imports instead of *

__version__ = '2025.6.0'  # YYYY.MM.MICRO  # MM not 0M, to match pip normalization.
# For non-release versions, use YYYY.MM.MICRO-dev,
#    to indicate the version has no corresponding release.
# in practical terms, to publish new release:
#    (1) remove '-dev' & update version number. E.g. '2024.12.0'
#    (2) push commit with message like: '[VERSION] 2024.12.0'
#    (3) publish release (e.g. git tag, flit build, flit publish)
#    (4) restore '-dev' but do not alter version number. E.g. '2024.12.0-dev'.
#    (5) push commit with message like: '[VERSION] 2024.12.0-dev'

from .addons import *
from .dimensions import *
from .hookups import *
from .mhd import *
from .multi_run_analysis import *
from .other_calculators import *
from .plotting import *
from .quantities import *
from .units import *
from .tools import *

from .defaults import DEFAULTS
from .errors import *
from .plasma_calculator import (
    DimensionlessPlasmaCalculator, ComponentHavingPlasmaCalculator,
    PlasmaCalculator, MultifluidPlasmaCalculator,
    VectorlessPlasmaCalculator, VectorlessMultifluidPlasmaCalculator,
)



"""
File Purpose: slices for fft.
"""
import warnings

from .fft_dimnames import FFTDimname
from ..docs_tools import format_docstring
from ..sentinels import UNSET
from ..xarray_tools import xarray_isel
from ...errors import InputError, DimensionKeyError

from ...defaults import DEFAULTS

_paramdocs_fft_slices = {
    'keep': f'''None, True, dict, or number in 0 < keep <= 1
        implies the fraction of each dimension to keep.
        (ignored for any dimensions which already have a slice specified.)
        e.g. keep=0.4 with length=1000 would result in slice(300, 700),
            since that keeps 400 out of 1000 points, and is centered at 1000/2.
        None --> ignored.
        True --> use keep = DEFAULTS.FFT_KEEP (default: {DEFAULTS.FFT_KEEP}).
        dict --> different value in each dimension;
                keys can be pre-fft OR post-fft dimension names.''',
    'half': '''None, str, or iterable of strs
        dimensions along which to keep only the right half of the result.
        (ignored for any dimensions which already have a slice specified.)
        None --> ignored.
        str or iterable of strs -->can be pre-fft OR post-fft dimension name(s).
        Applied after keep, e.g. keep=0.4, length=1000, half='x' --> slice(500, 700) for x.''',
    'step': '''None, dict, int, or non-integer between -1 and 1
        step to take along each dim.
        (ignored for any dimensions which already have a slice specified.)
        fractional value --> use fraction of length (e.g. 0.01 --> 1% of dim length), min |step|=1.
        negative --> reverses direction (and swaps start & stop for the slice)
        None --> equivalent to using step=1.
        dict --> different value in each dimension;
                keys can be pre-fft OR post-fft dimension names.''',
    'missing_slices': ''''ignore', 'warn', or 'raise'
        tells how to handle keys not matching any fft-related coordinate.
        'ignore' --> silently ignore these keys. This is the default.
        'warn' --> issue a warning.
        'raise' --> raise an error.'''
}


@format_docstring(**_paramdocs_fft_slices, sub_ntab=2)
class FFTSlices(dict):
    '''slicing fft result.
    keys should be names of dimensions and/or coordinates.
        pre-fft OR post-fft names both work (e.g. 'x', 'freq_x', or 'k_x')
    values should be slices, or int, iterable, or non-integer between -1 and 1.
        fractional indexing is supported as per interprets_fractional_indexing;
        non-integers between -1 and 1 are interpreted as fractions of the length,
            e.g. slice(0.25, 0.75) --> slice(int(0.25*L), int(0.75*L))

    slices: dict
        can provide keys and values here if desired, instead of directly as kwargs.

    The following can be provided during __init__ or set as attrs of self:
        SPECIAL SLICES OPTIONS:
            keep: {keep}
            half: {half}
            step: {step}
        OTHER OPTIONS:
            missing_slices: {missing_slices}
    '''
    # # # CREATION / INITIALIZATION # # #
    def __init__(self, *args_super, slices=dict(),
                 keep=None, half=None, step=None, missing_slices='ignore', **kw_super):
        super().__init__(*args_super, **slices, **kw_super)
        self.keep = keep
        self.half = half
        self.step = step
        self.missing_slices = missing_slices

    _kw_init = ('keep', 'half', 'step', 'missing_slices')

    def as_kw(self):
        '''return dict of kwargs to use to recreate this object.'''
        kw = {k: getattr(self, k) for k in self._kw_init}
        kw['slices'] = self
        return kw

    def copy(self):
        '''return a copy of self'''
        return type(self)(**self.as_kw())

    # # # APPLYING SLICES # # #
    @format_docstring(**_paramdocs_fft_slices, sub_ntab=1)
    def applied_to(self, fft_result, dims=None, *, missing_slices=UNSET):
        '''apply slices from self to fft_result.
        fft_result: array, probably xarray.DataArray.
            the result of an fft along one or more dimensions.
        dims: None or iterable of strs
            apply slicing only to these dimensions.
            None --> use fft_result.dims.
            any of these can be pre-fft dims, in which case will use the corresponding post-fft dim,
                inferred from fft_result. E.g. 'x' --> 'freq_x' or 'k_x' if one of those is in fft_result.
                Any pre-fft dims with no corresponding dim in fft_result will cause DimensionKeyError. 
        missing_slices: UNSET or {missing_slices}
            UNSET --> use self.missing_slices.
        additional kwargs are passed to self.get_slices.
        '''
        if dims is None:
            post_fft_dims = fft_result.dims
        else:
            post_fft_dims = []
            for d in dims:
                dname = FFTDimname.implied_from(d, fft_result.dims, post_fft=True)
                post_fft_dims.append(dname.post)
        slices = self.get_slices(post_fft_dims, missing_slices=missing_slices)
        return xarray_isel(fft_result, slices)

    @format_docstring(**_paramdocs_fft_slices, sub_ntab=1)
    def get_slices(self, post_fft_dims, *, missing_slices=UNSET):
        '''return slices for each dim in post_fft_dims.

        post_fft_dims: iterable of strs
            names of dimensions in the fft result, for which to get slices now.
        missing_slices: UNSET or {missing_slices}
            UNSET --> use self.missing_slices.

        result will be a dict with keys from post_fft_dims,
            excluding any keys for which self provides no slicing instructions.
        '''
        # missing slices bookkeeping
        if missing_slices is UNSET:
            missing_slices = self.missing_slices
        if missing_slices not in {'ignore', 'warn', 'raise'}:
            raise InputError(f"expected missing_slices in ('ignore', 'warn', 'raise'), but got {missing_slices}")
        if missing_slices != 'ignore':
            self._unfound_slices = set(self.keys())
        # getting result
        slices = dict()
        for dpost in post_fft_dims:
            dslice = self.get_slice(dpost)
            if dslice is not None:
                slices[dpost] = dslice
        # missing slices bookkeeping
        if missing_slices != 'ignore' and len(self._unfound_slices) > 0:
            errmsg = (f'Found slices keys {self._unfound_slices} not matching '
                      f'any post-fft dim ({post_fft_dims}) '
                      f'or pre-fft dim ({[FFTDimname.from_post(dpost).pre for dpost in post_fft_dims]}).')
            if missing_slices == 'warn':
                warnings.warn(errmsg)
            else:  # missing_slices == 'raise'
                raise DimensionKeyError(errmsg)
        # return result
        return slices

    def _mark_slice_as_found(self, key):
        '''self._unfound_slices -= {key}, if self._unfound_slices exists.'''
        if hasattr(self, '_unfound_slices'):
            self._unfound_slices -= {key}

    def get_slice(self, post_fft_dim):
        '''return slice for this dimension. Must provide name before & after fft,
        to ensure self is compatible with using pre & post fft names.
        '''
        if post_fft_dim in self:
            self._mark_slice_as_found(post_fft_dim)
            return self[post_fft_dim]
        # else, check pre_fft_dim
        pre_fft_dim = FFTDimname.from_post(post_fft_dim).pre
        if pre_fft_dim in self:
            self._mark_slice_as_found(pre_fft_dim)
            return self[pre_fft_dim]
        # else, get special slice
        return self.get_special_slice(post_fft_dim)
    
    def get_special_slice(self, post_fft_dim):
        '''return slice for this dim assuming it is a special case (only affected by keep, half, step).'''
        half = self.half
        keep = self.keep
        step = self.step
        if keep is None and half is None and step is None:
            return None
        # # else, get half, keep, & step along this dimension.
        pre_fft_dim = FFTDimname.from_post(post_fft_dim).pre
        # half
        half = [] if half is None else ([half] if isinstance(half, str) else half)
        half = pre_fft_dim in half or post_fft_dim in half
        # step
        if isinstance(step, dict):
            if pre_fft_dim in step:
                step = step[pre_fft_dim]
            elif post_fft_dim in step:
                step = step[post_fft_dim]
            else:
                step = None
        # keep
        input_keep = keep
        if isinstance(keep, dict):
            if pre_fft_dim in keep:
                keep = keep[pre_fft_dim]
            elif post_fft_dim in keep:
                keep = keep[post_fft_dim]
            else:
                keep = None
        if keep is True:
            keep = DEFAULTS.FFT_KEEP
        if keep is None:
            keep = 1
        if keep < 0 or keep > 1:
            raise InputError(f"expected keep in 0 < keep <= 1, but got keep={keep}")
        # calculate high & low values in slice, based on keep.
        if half:
            keeplow = 0.5
        else:
            keeplow = None if keep==1 else 0.5 - keep/2
        keephigh = None if keep==1 else 0.5 + keep/2
        # # now, get special slice
        if step is None or step > 0:
            return slice(keeplow, keephigh, step)
        else:
            return slice(keephigh, keeplow, step)

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''returns list of contents to go into self.__repr__.'''
        contents = []
        if self.keep is not None:
            contents.append(f'keep={self.keep!r}')
        if self.half is not None:
            contents.append(f'half={self.half!r}')
        if self.step is not None:
            contents.append(f'step={self.step!r}')
        if self.missing_slices != 'ignore':
            contents.append(f'missing_slices={self.missing_slices!r}')
        return contents

    def __repr__(self):
        contents = self._repr_contents()
        contents.append(super().__repr__())
        contents_str = ', '.join(contents)
        return f'{type(self).__name__}({contents_str})'

    # # # EQUALITY # # #
    def __eq__(self, other):
        '''return True if self is equal to other. (other can be any type.)'''
        if not isinstance(other, type(self)):
            return False
        if self.keep != other.keep:
            return False
        if self.half != other.half:
            return False
        if self.step != other.step:
            return False
        return super().__eq__(other)

    def __ne__(self, other):
        '''return self != other. Equivalent to: not (self==other).'''
        # (Defined in order to use __eq__ from this class instead of from super().)
        return not (self == other)



"""
File Purpose: indexing xarrays
"""
import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from .xarray_coords import xarray_index_coords
from .xarray_dimensions import (
    is_iterable_dim,
    xarray_ensure_dims, _paramdocs_ensure_dims,
    xarray_squeeze,
)
from ..arrays import interprets_fractional_indexing
from ..docs_tools import format_docstring, docstring_but_strip_examples
from ..os_tools import next_unique_name
from ..sentinels import UNSET
from ...errors import DimensionValueError


''' --------------------- Indexing --------------------- '''

@pcAccessor.register('isel')
@format_docstring(isel_doc=docstring_but_strip_examples(xr.DataArray.isel.__doc__),
                  fractional_indexing_doc=interprets_fractional_indexing.__doc__,
                  **_paramdocs_ensure_dims, sub_ntab=1)
def xarray_isel(array, indexers=None, *, promote_dims_if_needed=True,
                drop=False, missing_dims='raise', rounding='round', **indexers_as_kwargs):
    '''array.isel(...) which can also interpret fractional indexes between -1 and 1, and promotes non-dim coords.

    behaves just like xarray.DataArray.isel, but:
        - indexers also allow fractional indexes.
        - if any dim with index provided refers to a non-dimension coordinate, first promote it via swap_dims.
        - if any indexer has 'iseler' attr, use indexer.iseler(values) to determine indexes.
    In particular, for {{cname: index}}:
        - fractional indexes:
            if index is a slice, int, or iterable of ints, use it as is.
            if index contains any values between -1 and 1 (excluding -1, 0, and 1):
                treat that value as a fraction of L=len(array[cname]).
                E.g. 0.25 --> int(L * 0.25);
                    -0.1  --> -int(L * 0.1).
                This is equivalent to interprets_fractional_indexing(index, L)
        - non-dimension coordinates:
            if cname is a non-dimension coordinate, use xarray_promote_dim(array, cname).

    promote_dims_if_needed: {promote_dims_if_needed}
    drop, missing_dims: passed to array.isel; see below for details.
    rounding: passed to interprets_fractional_indexing; see below for details.

    xarray.DataArray.isel docs copied below:
    ----------------------------------------
        {isel_doc}

    interprets_fractional_indexing docs copied below:
    -------------------------------------------------
        {fractional_indexing_doc}
    '''
    if indexers is None:
        indexers = indexers_as_kwargs
    else:
        indexers = {**indexers, **indexers_as_kwargs}
    indexers_input = indexers
    array_input = array  # <-- helps with debugging in case of crash.
    # interpret fractional indexes, and promote coords to dims as necessary.
    indexers = dict()  # <-- not overwriting the originally-input value, this is a new dict.

    kw_ensure_dims = dict(promote_dims_if_needed=promote_dims_if_needed, missing_dims=missing_dims,
                          assert_1d=True,  # because here doesn't implement any way to index 2D+ dims.
                          return_existing_dims=True,  # so we can avoid indexing any missing dims!
                          )
    array, existing_dims = xarray_ensure_dims(array, list(indexers_input.keys()), **kw_ensure_dims)
    for cname in existing_dims:
        index = indexers_input[cname]
        coord = array.coords[cname]
        if hasattr(index, 'iseler'):
            ii = index.iseler(coord)
        else:
            ii = interprets_fractional_indexing(index, L=len(coord), rounding=rounding)
        indexers[cname] = ii
    # call isel
    return array.isel(indexers, drop=drop, missing_dims=missing_dims)

@pcAccessor.register('search')
def xarray_search(array, dim, value):
    '''return first index of value along dim
    (or coord. returns 0, not crash, if scalar coord which matches value.)
    Not efficient for large dims. For large sorted dims, see xarray.DataArray.searchsorted.

    crash with DimensionValueError if value not found in dim.
    '''
    for i, val in enumerate(np.atleast_1d(array.coords[dim].values)):
        if val == value:
            return i
    raise DimensionValueError(f'value={value!r} not found in array[{dim!r}]')

@pcAccessor.register('sel')
@format_docstring(sel_doc=docstring_but_strip_examples(xr.DataArray.sel.__doc__),
                  **_paramdocs_ensure_dims)
def xarray_sel(array, indexers=None, *, promote_dims_if_needed=True,
               method=None, tolerance=None, drop=False, **indexers_as_kwargs):
    '''array.sel(...) but prioritize general applicability over efficiency:
        - promote non-dimension coordinate dims first if applicable
        - (if coord.dtype is object) check coord equality,
            e.g. 0==Fluid('e', 0)=='e', so could use Fluid('e', 0), 'e', or 0 in sel.
            - can also use list, tuple, or 1D non-string iterable,
                e.g. ['e', 3, 'Fe+'] to get multiple fluids.
            - can also use slice,
                e.g. slice('e', 'Fe+', 2) to get every other fluid,
                starting from 'e', stopping before the first 'Fe+' match.
        - if indexer has 'iseler' attr, use indexer.iseler(values) to determine indexes.

    Assumes all indexing is for 1D dims. For indexing 2D+ dims, use xarray methods directly.

    promote_dims_if_needed: {promote_dims_if_needed}
    method: None or str
        method to use for inexact matches, for non-object dtype coords.

    xarray.DataArray.sel docs copied below:
    ----------------------------------------
        {sel_doc}
    '''
    if indexers is None:
        indexers = indexers_as_kwargs
    else:
        indexers = {**indexers, **indexers_as_kwargs}
    indexers_input = indexers
    sel_indexers = dict()  # indexing to delegate to xarray.sel
    obj_indexers = dict()  # indexing to handle here
    array_input = array  # <-- helps with debugging in case of crash.
    kw_ensure_dims = dict(promote_dims_if_needed=promote_dims_if_needed, missing_dims='raise',
                          assert_1d=True,  # because here doesn't implement any way to index 2D+ dims.
                          return_existing_dims=True,  # so we can avoid indexing any missing dims!
                          )
    array, existing_dims = xarray_ensure_dims(array, list(indexers_input.keys()), **kw_ensure_dims)
    for cname in existing_dims:
        if array[cname].dtype == object:
            obj_indexers[cname] = indexers_input[cname]
        else:
            sel_indexers[cname] = indexers_input[cname]
    # handle obj_indexers first.
    obj_isels = {}
    if len(obj_indexers) > 0:
        if method is not None:
            raise TypeError(f'cannot use method {method!r} with object dtype coords {list(obj_indexers)}.')
        for cname, index in obj_indexers.items():
            if hasattr(index, 'iseler'):
                obj_isels[cname] = index.iseler(array.coords[cname])
            elif is_iterable_dim(index):
                isel_here = []
                for ival in index:
                    isel_here.append(xarray_search(array, cname, ival))
                obj_isels[cname] = isel_here
            elif isinstance(index, slice):
                start, stop, step = index.start, index.stop, index.step
                istart, istop = None, None
                if start is not None:
                    istart = xarray_search(array, cname, start)
                if stop is not None:
                    istop = xarray_search(array, cname, stop)
                obj_isels[cname] = slice(istart, istop, step)
            else:  # index is a single value for a dim
                obj_isels[cname] = xarray_search(array, cname, index)
    array = array.isel(obj_isels, drop=drop)
    # handle sel_indexers
    return array.sel(sel_indexers, method=method, tolerance=tolerance, drop=drop)


''' --------------------- Custom where --------------------- '''

@pcAccessor.register('where')
def xarray_where(array, cond, other=UNSET, *, drop=False, skip_if_no_matching_dims=True):
    '''like xarray's builtin where, but return array unchanged if it has no dims matching cond.

    array: xarray.DataArray or xarray.Dataset
        array to apply condition to.
    cond: xarray.DataArray, xarray.Dataset, or callable
        Locations at which to preserve array's values. Must have dtype=bool.
        callable --> replace with cond(array).
    other: UNSET, scalar, DataArray, Dataset, or callable, optional
        Value to use for locations in array where cond is False.
        By default, these locations are filled with NA.
        callable --> replace with other(array).
        UNSET --> do not pass this arg to xarray.where()
    drop: bool, default: False
        If True, coordinate labels that only correspond to False values of
        the condition are dropped from the result.
    skip_if_no_matching_dims: bool, default: True
        If True, return array unchanged if it has no dims matching cond.
                if Dataset, keep data_vars unchanged if they have no dims matching cond.
        If False, return array.where(cond, other=other, drop=drop) directly.
    '''
    kw_where = dict(drop=drop) if other is UNSET else dict(other=other, drop=drop)
    if skip_if_no_matching_dims:  # custom processing here
        if callable(cond):
            cond = cond(array)
        if not any(d in array.coords for d in cond.dims):
            return array
        if isinstance(array, xr.Dataset):
            ds = array
            ds0 = ds  # ds before changing ds var.
            # hit data_vars with any dims matching cond, then append remaining vars.
            to_where = []
            to_skip = []
            for var, val in ds.data_vars.items():
                if any(d in val.coords for d in cond.dims):
                    to_where.append(var)
                else:
                    to_skip.append(var)
            if len(to_where) == 0:
                assert False, 'coding error, should have been handled above.'
            ds = ds[to_where].where(cond, **kw_where)
            if len(to_skip) > 0:
                ds = ds.assign(ds0[to_skip])
            return ds
    # else:
    return array.where(cond, **kw_where)


''' --------------------- Mapping --------------------- '''

@pcAccessor.register('map')
@format_docstring(**_paramdocs_ensure_dims)
def xarray_map(array, f, *args_f, axis=None, axes=None,
                     promote_dims_if_needed=True, missing_dims='raise', **kw_f):
    '''func(array, *args_f, **kw_f), but axis/axes can be provided as strings!

    Mainly useful if trying to apply f which expects unlabeled array & int axes inputs.
    E.g. numpy.mean can use axis kwarg as iterable of ints,
        but here can provide axis as a dim name str or list of dim names.
    Probably not super useful for mean, since xarray provides xr.mean,
        but may be useful for other functions e.g. scipy.ndimage.gaussian_filter,
        which might not have an existing equivalent in xarray.

    array: xarray.DataArray or Dataset
        apply f to this array, or each array in this Dataset
    f: callable
        will be called as f(array, *args_f, **kw_f),
        possibly will also be passed a value for axis or axes, if provided here
    axis, axes: None, str, or iterable of strs
        if provided, convert to axes positions in dataarray, and pass to f as int(s).
        Also promotes these coords to dims if necessary.
    promote_dims_if_needed: {promote_dims_if_needed}
    missing_dims: {missing_dims}
    '''
    # if Dataset, just apply this to each array.
    if isinstance(array, xr.Dataset):
        ds = array
        result = ds.copy()
        for name, arr in ds.data_vars.items():
            result[name] = xarray_map(arr, f, *args_f, axis=axis, axes=axes,
                                      promote_dims_if_needed=promote_dims_if_needed,
                                      missing_dims=missing_dims, **kw_f)
        return result
    # bookkeeping on 'axis' & 'axes' inputs:
    if axis is None and axes is None:  # simplest case; will just apply f to entire array.
        coords = None
        kw_ax = dict()
    elif axis is not None:  # and axes is None
        coords = axis
        ax_key = 'axis'
    elif axes is not None:  # and axis is None
        coords = axes
        ax_key = 'axes'
    else:  # both axis and axes were provided
        raise InputConflictError('cannot provide both "axis" and "axes".')
    if coords is not None:
        _single_ax_mode = isinstance(coords, str)
        if isinstance(coords, str):
            coords = [coords]
        # ensure the coords exist:
        array, existing_dims = xarray_ensure_dims(array, coords,
                                                  promote_dims_if_needed=promote_dims_if_needed,
                                                  missing_dims=missing_dims, return_existing_dims=True)
        if _single_ax_mode and len(existing_dims) == 1:
            existing_dims = existing_dims.pop()  # e.g., np.argsort expects int ax, not tuple!
        # convert coords to ax nums:
        ax_nums = array.get_axis_num(existing_dims)
        kw_ax = {ax_key: ax_nums}
    # call f but use xarray.Dataset.map functionality to preserve coords/attrs/etc.
    array_name = array.name
    _data_var_name = next_unique_name('_internal_variable', [*array.coords, *array.dims])
    ds = array.to_dataset(name=_data_var_name)
    ds_result = ds.map(f, args=args_f, **kw_f, **kw_ax)
    result = ds_result[_data_var_name].rename(array_name)
    return result


''' --------------------- Fancy sorting --------------------- '''

_paramdocs_sort = {
    # argsort & sort
    'newname': '''str
        result dim will be replaced by newname.format(dim=dim). Default: '{dim}_sort'.''',
    'ascending': '''bool
        result sorted in ascending order if True, descending if False.''',
    'nkeep': '''None or int
        number of sorted values to keep. E.g. nkeep=2 --> keep only the first 2 values.
        equivalent to using result.isel(newdim=(0, nkeep)), where newdim=newname.format(dim=dim).''',
    'promote_dims_if_needed': _paramdocs_ensure_dims['promote_dims_if_needed'],
    'squeeze': '''bool
        whether to drop redundant dims for argsort result.
        This means any dims with np.all(array==array.isel(dim=0)). (This includes dims with len=1.)''',
    'kind': '''None or str from {'quicksort', 'mergesort', 'heapsort', 'stable'}
        sorting algorithm. Passed directly to numpy.argsort; see help(numpy.argsort) for details.''',
    # sort, only
    'index': '''bool
        whether to use indexes for result's dim coord values, instead of original coord values.
        Either way, the relevant coord is multidimensional, varying across all array.dims except dim.
        True --> result['{dim}_index'] will be the indexes along original dim
        False --> result[dim] will be the original coord values.''',
    'newname_original': '''str
        name to use for original dim's coord values if store_original=True. Default: '{dim}_orig'.''',
}


@pcAccessor.register('argsort', totype='array')
@format_docstring(**_paramdocs_sort)
def xarray_argsort(array, dim, newname='{dim}_sort', *, ascending=True, nkeep=None,
                   promote_dims_if_needed=True, squeeze=True, kind=None):
    '''argsort for xarray.DataArray; returns indexes along dim which would sort array along dim.
    Drops original coords along dim, because the output along dim is now the indexes;
        result[i] corresponds to coord value array[dim][result[i]], not array[dim][i].

    Good examples to consider (where result=xarray_argsort(array, dim)):
        (1) sortarr = array.isel(dim=result) to get sorted array,
            which will have dim as a multidimensional coordinate with same dims as argsort result,
            and values telling original values of dim, sorted appropriately.
        (2) sortarr = array.isel(dim=result.isel(newdim=slice(0, N)) to get sorted array,
            similar to example (1), but only keeping first N lowest values
            (or N highest values if ascending=False when getting argsort result).

    array: xarray.DataArray
        array to argsort.
    dim: str
        dimension along which to sort.
    newname: {newname}
    ascending: {ascending}
    nkeep: {nkeep}
    promote_dims_if_needed: {promote_dims_if_needed}
    squeeze: {squeeze}
    kind: {kind}
    '''
    result = xarray_map(array, np.argsort, axis=dim,
                        promote_dims_if_needed=promote_dims_if_needed,
                        kind=kind)
    if not ascending:
        result = result.isel({dim: slice(None, None, -1)})
    newdim = newname.format(dim=dim)
    if newdim in array:
        raise NotImplementedError(f'xarray_argsort when newdim={newdim!r} already exists in array')
    result = result.drop_vars(dim).rename({dim: newdim})
    if nkeep is not None:
        result = result.isel({newdim: slice(0, nkeep)})
    if squeeze:
        result = xarray_squeeze(result)
    return result

@pcAccessor.register('sort_along', totype='array')
@format_docstring(**_paramdocs_sort)
def xarray_sort_array_along(array, dim, newname='{dim}_sort', *, ascending=True, nkeep=None,
                            index=False, store_original=False, newname_original='{dim}_orig',
                            promote_dims_if_needed=True, squeeze_argsort=True, kind=None):
    '''returns array, sorted along dim.
    
    array: xarray.DataArray
        array to sort.
    dim: str
        dimension along which to sort.
    newname: {newname}
    ascending: {ascending}
    nkeep: {nkeep}
    index: {index}
    store_original: bool
        whether to store original dim's coord values in result.
        True --> result will be a Dataset with data_vars array.name and '{{dim}}_orig';
            result[array.name] = sorted values (dims=array.dims but replace dim with newname.format(dim=dim))
            result['{{dim}}_orig'] = original values of dim (dims='{{dim}}_orig')
        False --> result will be a DataArray which does not track original dim's coord values separately.
    newname_original: {newname_original}
    promote_dims_if_needed: {promote_dims_if_needed}
    squeeze_argsort: {squeeze}
    kind: {kind}
    '''
    if store_original and array.name is None:
        raise InputConflictError('cannot use store_original=True, when array.name is None.')
    asort = xarray_argsort(array, dim, newname=newname, ascending=ascending, nkeep=nkeep,
                            promote_dims_if_needed=promote_dims_if_needed,
                            squeeze=squeeze_argsort, kind=kind)
    array_input = array
    if index:
        array = xarray_index_coords(array, dim, drop=True)
    result = array.isel({dim: asort})
    if store_original:
        origname = newname_original.format(dim=dim)
        if origname in array_input:
            raise NotImplementedError(f'xarray_sort_array_along when origname={origname!r} already exists')
        original = array_input[dim].rename({dim: origname})
        result = xr.Dataset({array.name: result, origname: original})
    return result

@pcAccessor.register('sort_along', totype='dataset')
@format_docstring(**_paramdocs_sort)
def xarray_sort_dataset_along(ds, var, dim, newname='{dim}_sort', *, ascending=True, nkeep=None,
                              index=False, store_original=True, newname_original='{dim}_orig',
                              promote_dims_if_needed=True, squeeze_argsort=True, kind=None):
    '''returns dataset, sorted along var's values for dim.

    dataset: xarray.Dataset
        dataset to sort.
    var: str
        data_var which determines sort order.
    dim: str
        dimension along which to sort.
    newname: {newname}
    ascending: {ascending}
    nkeep: {nkeep}
    index: {index}
    store_original: bool
        whether to store original dim's coord values in result.
        True --> result['{{dim}}_orig'] = original values of dim (dims='{{dim}}_orig')
        False --> do not add original values of dim as a data_var in result.
    newname_original: {newname_original}
    promote_dims_if_needed: {promote_dims_if_needed}
    squeeze_argsort: {squeeze}
    kind: {kind}
    '''
    ds_input = ds
    array = ds[var]
    asort = xarray_argsort(array, dim, newname=newname, ascending=ascending, nkeep=nkeep,
                            promote_dims_if_needed=promote_dims_if_needed,
                            squeeze=squeeze_argsort, kind=kind)
    if index:
        ds = xarray_index_coords(ds, dim, drop=True)
    result = ds.isel({dim: asort})
    if store_original:
        origname = newname_original.format(dim=dim)
        if origname in ds_input:
            raise NotImplementedError(f'xarray_sort_dataset_along when origname={origname!r} already exists')
        original = ds_input[dim].rename({dim: origname})
        result[origname] = original
    return result


''' --------------------- Fancy argmin/max --------------------- '''
# [EFF] uses np.argmin/max instead of reusing algorithms above for a full sort.
# For argmin/max itself, use xarray builtin, e.g. array.argmin(dim=dim).

@pcAccessor.register('cmin', totype='array')
@format_docstring(**_paramdocs_sort)
def xarray_cmin(array, coord, where=UNSET, *, promote_dims_if_needed=True):
    '''return (array of) coord value(s) at array.argmin(dim=dim associated with coord).
    coord must be a dimension, or associated with a single 1D dimension.

    array: xarray.DataArray
        array to find minimum value in.
    coord: str
        coord along which to find min.
    promote_dims_if_needed: {promote_dims_if_needed}

    see also: xarray_min_coord_where
    '''
    array = xarray_ensure_dims(array, coord, promote_dims_if_needed=promote_dims_if_needed)
    argmin = array.argmin(coord)
    return array[coord].isel({coord: argmin})

@pcAccessor.register('cmax', totype='array')
@format_docstring(**_paramdocs_sort)
def xarray_cmax(array, coord, where=UNSET, *, promote_dims_if_needed=True):
    '''return (array of) coord value(s) at array.argmax(dim=dim associated with coord).
    coord must be a dimension, or associated with a single 1D dimension.

    array: xarray.DataArray
        array to find max value in.
    coord: str
        coord along which to find max.
    promote_dims_if_needed: {promote_dims_if_needed}

    see also: xarray_max_coord_where
    '''
    array = xarray_ensure_dims(array, coord, promote_dims_if_needed=promote_dims_if_needed)
    argmax = array.argmax(coord)
    return array[coord].isel({coord: argmax})

@pcAccessor.register('varmin', totype='dataset')
@format_docstring(**_paramdocs_sort)
def xarray_varmin(ds, *, promote_dims_if_needed=True):
    '''return (array of) var name(s) for var with minimum values in ds.
    equivalent: xarray_cmin(ds.to_dataarray(), 'variable')

    ds: xarray.Dataset
        dataset to find minimum value in.
    promote_dims_if_needed: {promote_dims_if_needed}
    '''
    return xarray_cmin(ds.to_array(), 'variable', promote_dims_if_needed=promote_dims_if_needed)

@pcAccessor.register('varmax', totype='dataset')
@format_docstring(**_paramdocs_sort)
def xarray_varmax(ds, *, promote_dims_if_needed=True):
    '''return (array of) var name(s) for var with maximum values in ds.
    equivalent: xarray_cmax(ds.to_dataarray(), 'variable')

    ds: xarray.Dataset
        dataset to find maximum value in.
    promote_dims_if_needed: {promote_dims_if_needed}
    '''
    return xarray_cmax(ds.to_array(), 'variable', promote_dims_if_needed=promote_dims_if_needed)

@pcAccessor.register('at_min_of')
@format_docstring(**_paramdocs_sort)
def xarray_at_min_of(array, of, dim=None, *, promote_dims_if_needed=True):
    '''return array values at locations of minimum `of`. Roughly: array.isel(of.argmin([dim])).
    But, here is nicer, in a few ways:
        (1) here doesn't require array to have all dims in dim.
            E.g., if array has dims 'x', 'kmod', and `of` has dims 'kmod', 'kang',
                and dim=['kmod', 'kang'], it will be handled in the intuitive way. Roughly:
                    amin = of.argmin(['kmod', 'kang'])
                    result = array.isel(kmod=amin['kmod'])
        (2) here allows All-NaN slice in `of`.
            all-nan slices in `of` will be filled with nan values in result.
            E.g., if dims=['d0', 'd1'], `of` has dims ['d0', 'd1', 'nmul', 'fluid'],
                and of.isnull().all(['d0', 'd1']) when nmul index == 4 or 5, and fluid index == 2,
                then result.isel(nmul=[4,5], fluid=2) will be filled with nan values.

    array: xarray.DataArray or xarray.Dataset
        array to get result values from.
    of: xarray.DataArray
        array to find minimum values in.
    dim: None, str, or list of str
        dim(s) along which to find the minimum values of `of`.
        Takes the "simultaneous" minimum along all these dims, as per argmin(dims)
        None --> use of.dims.
        (if dim is a list with len==0, return array, unchanged.)
    promote_dims_if_needed: {promote_dims_if_needed}
    '''
    if dim is None:
        dim = of.dims
    if isinstance(dim, str):
        dim = [dim]
    if len(dim) == 0:
        return array
    # compute indexing
    allnan = of.isnull().all(dim)
    any_allnan = allnan.any()
    if any_allnan:
        of = of.fillna(np.inf)  # np.inf never interferes with argmin results.
    idx = of.argmin(dim)
    # apply indexing to array
    array, existing = xarray_ensure_dims(array, dim, promote_dims_if_needed=promote_dims_if_needed,
                                         missing_dims='ignore', return_existing_dims=True)
    result = array.isel({d: idx[d] for d in existing})
    if any_allnan:
        result = result.where(~allnan)
    return result

@pcAccessor.register('at_max_of')
@format_docstring(**_paramdocs_sort)
def xarray_at_max_of(array, of, dim=None, *, promote_dims_if_needed=True):
    '''return array values at locations of maximum `of`. Roughly: array.isel(of.argmax([dim])).
    But, here is nicer, in a few ways:
        (1) here doesn't require array to have all dims in dim.
            E.g., if array has dims 'x', 'kmod', and `of` has dims 'kmod', 'kang',
                and dim=['kmod', 'kang'], it will be handled in the intuitive way. Roughly:
                    amax = of.argmax(['kmod', 'kang'])
                    result = array.isel(kmod=amax['kmod'])
        (2) here allows All-NaN slice in `of`.
            all-nan slices in `of` will be filled with nan values in result.
            E.g., if dims=['d0', 'd1'], `of` has dims ['d0', 'd1', 'nmul', 'fluid'],
                and of.isnull().all(['d0', 'd1']) when nmul index == 4 or 5, and fluid index == 2,
                then result.isel(nmul=[4,5], fluid=2) will be filled with nan values.

    array: xarray.DataArray or xarray.Dataset
        array to get result values from.
    of: xarray.DataArray
        array to find maximum values in.
    dim: None, str, or list of str
        dim(s) along which to find the maximum values of `of`.
        Takes the "simultaneous" maximum along all these dims, as per argmax(dims)
        None --> use of.dims.
        (if dim is a list with len==0, return array, unchanged.)
    promote_dims_if_needed: {promote_dims_if_needed}
    '''
    # [TODO] encapsulate repeated code from xarray_at_min_of.
    if dim is None:
        dim = of.dims
    if isinstance(dim, str):
        dim = [dim]
    if len(dim) == 0:  # edge case
        return array
    # compute indexing
    allnan = of.isnull().all(dim)  # check for allnan slices
    any_allnan = allnan.any()
    if any_allnan:
        of = of.fillna(-np.inf)  # -np.inf never interferes with argmax results.
    idx = of.argmax(dim)
    # apply indexing to array
    array, existing = xarray_ensure_dims(array, dim, promote_dims_if_needed=promote_dims_if_needed,
                                         missing_dims='ignore', return_existing_dims=True)
    result = array.isel({d: idx[d] for d in existing})
    if any_allnan:
        result = result.where(~allnan)
    return result

@pcAccessor.register('min_coord_where', totype='array')
@format_docstring(**_paramdocs_sort)
def xarray_min_coord_where(array, coord, where,  *, promote_dims_if_needed=True):
    '''return (array of) minimum value(s) of coord, where condition is True.

    array: xarray.DataArray
        array to find minimum value in.
    coord: str
        coord whose minimum value will appear in the result.
        Must correspond with a single dimension of array.
    where: xarray.DataArray or callable
        locations at which to consider values of coord.
        callable --> use where=where(array)
                    (will promote coord to dim first, if coord not dim yet.)
    promote_dims_if_needed: {promote_dims_if_needed}

    Example:
        xarray_min_coord_where(growth, 'E', lambda arr: arr>0)
        # returns the minimum value(s) of E across all regions where growth>0.
        # (result has no 'E' dim, but does retain any other dims from array.)

        Compare to xarray_cmin(growth.where(growth>0), 'E'),
        which tells the value of E at the location of minimum growth (where growth>0)
        (but not necessarily the minimum E across all regions with growth>0).
    '''
    array = xarray_ensure_dims(array, coord, promote_dims_if_needed=promote_dims_if_needed)
    if callable(where):
        where = where(array)
    tmp = array[coord].where(where)
    return tmp.min(dim=coord)

@pcAccessor.register('max_coord_where', totype='array')
@format_docstring(**_paramdocs_sort)
def xarray_max_coord_where(array, coord, where,  *, promote_dims_if_needed=True):
    '''return (array of) maximum value(s) of coord, where condition is True.

    array: xarray.DataArray
        array to find maximum value in.
    coord: str
        coord whose maximum value will appear in the result.
        Must correspond with a single dimension of array.
    where: xarray.DataArray or callable
        locations at which to consider values of coord.
        callable --> use where=where(array)
                    (will promote coord to dim first, if coord not dim yet.)
    promote_dims_if_needed: {promote_dims_if_needed}

    Example:
        xarray_max_coord_where(growth, 'kmod', lambda arr: arr>0)
        # returns the maximum value(s) of kmod across all regions where growth>0.
        # (result has no 'kmod' dim, but does retain any other dims from array.)

        Compare to xarray_cmax(growth.where(growth>0), 'kmod'),
        which tells the value of kmod at the location of max growth (where growth>0)
        (but not necessarily the maximum kmod across all regions with growth>0).
    '''
    array = xarray_ensure_dims(array, coord, promote_dims_if_needed=promote_dims_if_needed)
    if callable(where):
        where = where(array)
    tmp = array[coord].where(where)
    return tmp.max(dim=coord)



"""
File Purpose: display / print objects
"""

import pydoc


''' --------------------- simple repr --------------------- '''

def repr_simple(obj):
    '''return obj._repr_simple() if possible, else repr(obj)'''
    try:
        obj_repr_simple = obj._repr_simple
    except AttributeError:
        return repr(obj)
    else:
        return obj_repr_simple()


''' --------------------- Misc. --------------------- '''

def print_clear(N=80):
    '''clears current printed line of up to N characters, and returns cursor to beginning of the line.
    debugging: make sure to use print(..., end=''), else your print statement will go to the next line.
    '''
    print('\r'+ ' '*N +'\r',end='')

def help_str(f):
    '''gets string from help(f)'''
    return pydoc.render_doc(f, '%s')


''' --------------------- Pretty str formatting --------------------- '''

def join_strs_with_max_line_len(strs, sep=', ', max=80, *, key=len):
    '''join strs with sep, with max_line_len characters per line;
    if joining another str in the same line makes line too long, start new line instead.
    if sep ends with space and appears at end of line, rstrip() that line.

    max: int or None, default 80
        maximum length of line. None -> no maximum (equivalent to sep.join(strs)).
    key: callable, default len
        key(str) should return length associated with str.
    '''
    if max is None:
        return sep.join(strs)
    if len(strs)==0:
        return ''
    lines = []
    line = strs[0]  # line currently being built
    L = key(strs[0])  # length of line being built.
    for r in strs[1:]:
        rlen = key(r)
        len_if_added = L + len(sep) + rlen
        if (len_if_added <= max):  # r goes on this line.
            line = f'{line}{sep}{r}'
            L = len_if_added
        else:
            lines.append(f'{line}{sep.rstrip()}')
            line = r
            L = rlen
    # finalize last line
    lines.append(line)
    # join lines
    result = '\n'.join(lines)
    return result
    


"""
File Purpose: itAccessor allows to put array.it.{method} options in xarray objects.

This file also puts wavevector dims-related attrs in itAccessor.
"""

from ...tools import XarrayAccessor


''' --------------------- it accessor --------------------- '''
# it can stand for "instability theory" or "instability tools" :)

class itAccessor(XarrayAccessor, accessor_name='it', access_type=None):
    '''Accessor for instability theory tools.'''
    pass  # attach methods & attrs via itAccessor.register & register_attr.

class itArrayAccessor(itAccessor, access_type='array'):
    '''Accessor for instability theory tools on DataArrays.'''
    pass  # attach methods & attrs via itAccessor.register & register_attr with totype='array'.

class itDatasetAccessor(itAccessor, access_type='dataset'):
    '''Accessor for instability theory tools on Datasets.'''
    pass  # attach methods & attrs via itAccessor.register & register_attr with totype='dataset'.



"""
File Purpose: UnitsManagerPIC class
"""

from .units_manager import UnitsManager, _units_manager_helpstr
from ..errors import InputMissingError, InputConflictError
from ..tools import format_docstring
from ..defaults import DEFAULTS


''' --------------------- UnitsManagerPIC --------------------- '''

@format_docstring(helpstr=_units_manager_helpstr, sub_ntab=1)
class UnitsManagerPIC(UnitsManager):
    '''units manager with the from_pic method, and some other helpful PIC things.

    Manages units, including conversion factors and physical constants in different unit systems.

    --- UnitsManagerPIC.help() will print a helpful message: ---
        {helpstr}
    '''
    SIC_QUANTITIES = ("time", "length", "mass", "current", "temperature", "charge", "frequency", "speed", "wave_number",
                "mass_density", "charge_density", "number_density", "current_density", "permittivity", "permeability",
                "pressure", "energy", "power", "potential", "b_field", "h_field", "e_field")
    # ^ SIC stands for self.sic here, which stores "si conversions". It's not a typo from 'PIC'.
    _PIC_DEFAULTS = {key: UnitsManager.PHYSICAL_CONSTANTS_SI[key][0] for key in ('qe', 'me', 'eps0')}

    @classmethod
    def from_pic(cls, units='si', *,
                 u_l=None, u_t=None, u_n=None, ne=None, ne_si=None,
                 qe=None, me=None, eps0=None,
                 qe_si=None, me_si=None, eps0_si=None,
                 M=None, q=None, u_eps=None,
                 **kw_init):
        """create a UnitsManagerPIC from pic parameters.
        'raw' units refers to the units from the PIC code.

        REQUIRED PARAMETERS
            There is an ambiguity in the units from a PIC code (e.g. from eppic).
            Indicate length or time units in one of the following ways:
                u_l: length [si] = u_l * length [raw]
                u_t: time [si] = u_t * time [raw]
                u_n: number density [si] = u_n * number density [raw]
                ne and ne_si: number density n for electrons; ne = n [raw]; ne_si = n [si]

        OPTIONAL PARAMETERS
            :param qe: None or value
                |electron charge| [raw units].
                if None, use cls._PIC_DEFAULTS['qe']. Default: SI value of qe.
            :param me: None or value
                electron mass [raw units].
                if None, use cls._PIC_DEFAULTS['me']. Default: SI value of me.
            :param eps0: None or value
                permittivity of free space [raw units].
                if None, use cls._PIC_DEFAULTS['eps0']. Default: SI value of epsilon0
            :param qe_si, me_si, eps0_si: None or value
                optionally, specify a non-traditional SI value
                for electron charge, electron mass, and/or epsilon0.
            :param M, q, u_eps: None or value
                if provided, set u_M = M, u_q = q, or u_eps=u_eps,
                and ignore info about me, qe, or eps0.
        """
        # get required value
        u_l_and_t = cls.pic_ambiguous_unit(u_l=u_l, u_t=u_t, u_n=u_n, ne=ne, ne_si=ne_si)
        u_l = u_l_and_t['u_l']
        u_t = u_l_and_t['u_t']

        # get defaults / optional kwargs values
        if qe is None: qe = cls._PIC_DEFAULTS['qe']
        if me is None: me = cls._PIC_DEFAULTS['me']
        if eps0 is None: eps0 = cls._PIC_DEFAULTS['eps0']

        if qe_si is None: qe_si = cls.PHYSICAL_CONSTANTS_SI['qe'][0]
        if me_si is None: me_si = cls.PHYSICAL_CONSTANTS_SI['me'][0]
        if eps0_si is None: eps0_si = cls.PHYSICAL_CONSTANTS_SI['eps0'][0]

        # GET BASE RAW UNITS (CHARGE, MASS, TEMPERATURE, LENGTH, TIME)
        u_q = qe_si / qe if q is None else q  # qe [raw] * u_q == qe [si]
        u_M = me_si / me if M is None else M  # me [raw] * u_M == me [si]
        u_eps0 = eps0_si / eps0 if u_eps is None else u_eps  # eps0 [raw] * u_eps0 == eps0 [si]
        # eps0 as units constraint:
        #   F = (1 / (4 pi eps0)) * (q1 q2 / r^2)
        #   --> u_M * u_l * u_t^-2 = u_eps0^-1 * u_q^2 * u_l^-2
        #   --> u_l^3 * u_t^-2 = u_M^-1 * u_eps0^-1 * u_q^2
        #   --> u_l = (u_M^-1 * u_eps0^-1 * u_q^2 * u_t^2)**(1/3)
        #       u_t = (u_M * u_eps0 * u_q^-2 * u_l^3)**(1/2)
        if u_l is None:  # u_t is known
            u_l = (u_M**-1 * u_eps0**-1 * u_q**2 * u_t**2)**(1/3)
        else:   # u_l is known
            u_t = (u_M * u_eps0 * u_q**-2 * u_l**3)**(1/2)

        # return
        return cls(units=units, q=u_q, M=u_M, l=u_l, t=u_t, **kw_init)

    @staticmethod
    def pic_ambiguous_unit(u_l=None, u_t=None, u_n=None, ne=None, ne_si=None):
        '''Returns dict of u_l and u_t.
        Requires that u_l or u_t can be determined, but not both.
        raise InputConflictError if too many inputs; InputMissingError if not enough inputs.

        Indicate length or time units in one of the following ways:
            u_l: length [si] = u_l * length [raw]
            u_t: time [si] = u_t * time [raw]
            u_n: number density [si] = u_n * number density [raw]
            ne and ne_si: number density n for electrons; ne = n [raw]; ne_si = n [si]

        DEFAULTS.pic_ambiguous_unit can be set to a dict of default values instead;
            if DEFAULTS.pic_ambiguous_unit exists, use it when all inputs to this function are None.
            E.g. DEFAULTS.pic_ambiguous_unit = dict(u_t=1) --> pic seconds are SI seconds, by default.
            Ignored when any input is not None (or when providing ne or ne_si but not both).
        '''
        n_provided = sum((val is not None) for val in (u_l, u_t, u_n, ne and ne_si))
        if n_provided < 1:
            default = getattr(DEFAULTS, 'pic_ambiguous_unit', dict())
            if len(default) == 0:
                errmsg = "UnitsManagerPIC requires one of u_l, u_t, u_n, or (ne and ne_si)"
                raise InputMissingError(errmsg)
            else:
                n_provided = len(default)
                u_l = default.get('u_l', None)
                u_t = default.get('u_t', None)
                u_n = default.get('u_n', None)
                ne = default.get('ne', None)
                ne_si = default.get('ne_si', None)
        if n_provided > 1:
            errmsg = "UnitsManagerPIC requires only one of u_l, u_t, u_n, or (ne and ne_si)"
            raise InputConflictError(errmsg)
        if (ne is not None) and (ne_si is not None):
            u_n = ne / ne_si
        if u_n is not None:
            u_l = u_n ** (-1/3)
        return dict(u_l=u_l, u_t=u_t)
